#include "hell.h"

#include "hell-x.hh"
#include "expansionSFs.hh"
#include <LHAPDF/LHAPDF.h>

#include "pmom.h"
#include "mellinpdf.h"
#include "settings.h"
#include "mellinint.h"
#include "scales.h"
#include "phasespace.h"
#include "mesq.h"
#include "resint.h"
#include "pdfevol.h"
#include "resconst.h"
#include "constants.h"
#include "psi.h"
#include "evolnum.h"
#include "evoltrunc.h"
#include "pdf.h"
#include "gaussrules.h"

using namespace constants;
using namespace parton;

int                       hell::nf;
HELLx::Order              hell::order;
HELLx::LogOrder           hell::logorder;
double                   *hell::alphas;
int                       hell::nas;
map <int,HELLx::xTableP*> hell::xT;
HELLx::HELLxnf           *hell::sxtable;

int hell::dim;

complex <double> *hell::gli;
complex <double> *hell::vai;
complex <double> *hell::m3i;
complex <double> *hell::m8i;
complex <double> *hell::m15i;
complex <double> *hell::m24i;
complex <double> *hell::sgi;
complex <double> *hell::p3i;
complex <double> *hell::p8i;
complex <double> *hell::p15i;
complex <double> *hell::p24i;
complex <double> *hell::cxn;
double *hell::xdpgg;
double *hell::xdpqg;
double *hell::xdpgg_1;
double *hell::xdpqg_1;
double *hell::xdpgg_2;
double *hell::xdpqg_2;

cmatrix *hell::usg;

double hell::beta0(int nf) { return (33.-2.*nf)/12./M_PI; }
double hell::beta1(int nf) { return (153.-19.*nf)/24./M_PI/M_PI; }
double hell::beta2(int nf) { return (2857. - 5033./9.*nf + 325./27.*nf*nf)/128./M_PI/M_PI/M_PI; }
double hell::beta3(int nf) { return (149753./6. + 3564.*zeta3 + nf*(-1078361./162.-6508./27.*zeta3)+nf*nf*(50065./162.+6472./81.*zeta3)+1093./729.*nf*nf*nf)/256./M_PI/M_PI/M_PI/M_PI; }

void hell::smallx_update(double x, double* xdPgg, double *xdPqg)
{
  for(int i = 0; i < nas; i++)
    hell::xT[i]->eval(x,xdPgg[i],xdPqg[i]);
}

void hell::smallx_eval(double x, int i, double as, double fsx[4])
{
  if(as < hell::alphas[0]/2. || as > 4.*hell::alphas[hell::nas-1])
    cout << "HELLx: WARNING: alpha_s=" << as << " out of interpolation range [" << hell::alphas[0] << ", " << hell::alphas[hell::nas-1] << "] for nf=" << hell::nf << endl;

  //k is the index of the closest alphas grid
  int k = 0;
  for (unsigned int i = 1; i < hell::nas; i++)
    {
      if (as < hell::alphas[i]) break;
      k++;
    }
  // the following is needed for cubic interpolation
  if (k==0) k++;
  if (k==hell::nas-1) k -= 2;
  if (k==hell::nas-2) k -= 1;
  
  double xdPgg[hell::nas], xdPqg[hell::nas];
  copy (hell::xdpgg+i*hell::nas, hell::xdpgg+(i+1)*hell::nas, xdPgg);
  copy (hell::xdpqg+i*hell::nas, hell::xdpqg+(i+1)*hell::nas, xdPqg);
  
  //cubic interpolation
  double dGgg =(
    +(as-alphas[k  ])*(as-alphas[k+1])*(as-alphas[k+2])*xdPgg[k-1]/(alphas[k-1]-alphas[k  ])/(alphas[k-1]-alphas[k+1])/(alphas[k-1]-alphas[k+2])
    +(as-alphas[k-1])*(as-alphas[k+1])*(as-alphas[k+2])*xdPgg[k  ]/(alphas[k  ]-alphas[k-1])/(alphas[k  ]-alphas[k+1])/(alphas[k  ]-alphas[k+2])
    +(as-alphas[k-1])*(as-alphas[k  ])*(as-alphas[k+2])*xdPgg[k+1]/(alphas[k+1]-alphas[k-1])/(alphas[k+1]-alphas[k  ])/(alphas[k+1]-alphas[k+2])
    +(as-alphas[k-1])*(as-alphas[k  ])*(as-alphas[k+1])*xdPgg[k+2]/(alphas[k+2]-alphas[k-1])/(alphas[k+2]-alphas[k  ])/(alphas[k+2]-alphas[k+1])
		)/x;
  double dGqg =(
    +(as-alphas[k  ])*(as-alphas[k+1])*(as-alphas[k+2])*xdPqg[k-1]/(alphas[k-1]-alphas[k  ])/(alphas[k-1]-alphas[k+1])/(alphas[k-1]-alphas[k+2])
    +(as-alphas[k-1])*(as-alphas[k+1])*(as-alphas[k+2])*xdPqg[k  ]/(alphas[k  ]-alphas[k-1])/(alphas[k  ]-alphas[k+1])/(alphas[k  ]-alphas[k+2])
    +(as-alphas[k-1])*(as-alphas[k  ])*(as-alphas[k+2])*xdPqg[k+1]/(alphas[k+1]-alphas[k-1])/(alphas[k+1]-alphas[k  ])/(alphas[k+1]-alphas[k+2])
    +(as-alphas[k-1])*(as-alphas[k  ])*(as-alphas[k+1])*xdPqg[k+2]/(alphas[k+2]-alphas[k-1])/(alphas[k+2]-alphas[k  ])/(alphas[k+2]-alphas[k+1])
		)/x;

  
  bool useLLp = false; //-->Make this an option
  int RCvar = 0;       //-->Make this an option
  if (opts.order_evol == 1)     //LO+LLx
    dGgg += as*as * ( HELLx::PLL1(x,hell::nf) - HELLx::mcPgg1LL(x,hell::nf) );
  else if(opts.order_evol == 2) //NLO+NLLx
    {
      dGgg += as*as*as * ( HELLx::PNLL2(x,hell::nf) - CF/CA*HELLx::Pqg2(x,hell::nf,useLLp) - HELLx::mcPgg2NLL(x,hell::nf,useLLp) );
      dGqg += as*as*as *   HELLx::Pqg2(x,hell::nf,useLLp);
    }
  //else if(opts.order_evol == 3) //NNLO+NLLx
    //dGqg += as*as*as*as *   HELLx::Pqg3(x,hell::nf,RCvar);
  else if(opts.order_evol == 4)   //N3LO+NLLx
    {
      dGgg -= as*as*as*as * ( HELLx::PNLL3(x,hell::nf,RCvar) - CF/CA*HELLx::Pqg3(x,hell::nf,useLLp,RCvar) - HELLx::mcPgg3NLL(x,hell::nf,useLLp,RCvar) );
      dGqg -= as*as*as*as *   HELLx::Pqg3(x,hell::nf,useLLp,RCvar);
    }

  HELLx::sqmatrix<double> sg(dGgg, CF/CA*dGgg, dGqg, CF/CA*dGqg);

  fsx[0] = sg.gg();
  fsx[1] = sg.gq();
  fsx[2] = sg.qg();
  fsx[3] = sg.qq();
}

void hell::smallx_eval_1(double x, int i, double as, double fsx[4])
{
  if(as < hell::alphas[0]/2. || as > 4.*hell::alphas[hell::nas-1])
    cout << "HELLx: WARNING: alpha_s=" << as << " out of interpolation range [" << hell::alphas[0] << ", " << hell::alphas[hell::nas-1] << "] for nf=" << hell::nf << endl;

  //k is the index of the closest alphas grid
  int k = 0;
  for (unsigned int i = 1; i < hell::nas; i++)
    {
      if (as < hell::alphas[i]) break;
      k++;
    }
  // the following is needed for cubic interpolation
  if (k==0) k++;
  if (k==hell::nas-1) k -= 2;
  if (k==hell::nas-2) k -= 1;
  
  double xdPgg[hell::nas], xdPqg[hell::nas];
  copy (hell::xdpgg_1+i*hell::nas, hell::xdpgg_1+(i+1)*hell::nas, xdPgg);
  copy (hell::xdpqg_1+i*hell::nas, hell::xdpqg_1+(i+1)*hell::nas, xdPqg);
  
  //cubic interpolation
  double dGgg =(
    +(as-alphas[k  ])*(as-alphas[k+1])*(as-alphas[k+2])*xdPgg[k-1]/(alphas[k-1]-alphas[k  ])/(alphas[k-1]-alphas[k+1])/(alphas[k-1]-alphas[k+2])
    +(as-alphas[k-1])*(as-alphas[k+1])*(as-alphas[k+2])*xdPgg[k  ]/(alphas[k  ]-alphas[k-1])/(alphas[k  ]-alphas[k+1])/(alphas[k  ]-alphas[k+2])
    +(as-alphas[k-1])*(as-alphas[k  ])*(as-alphas[k+2])*xdPgg[k+1]/(alphas[k+1]-alphas[k-1])/(alphas[k+1]-alphas[k  ])/(alphas[k+1]-alphas[k+2])
    +(as-alphas[k-1])*(as-alphas[k  ])*(as-alphas[k+1])*xdPgg[k+2]/(alphas[k+2]-alphas[k-1])/(alphas[k+2]-alphas[k  ])/(alphas[k+2]-alphas[k+1])
		)/x;
  double dGqg =(
    +(as-alphas[k  ])*(as-alphas[k+1])*(as-alphas[k+2])*xdPqg[k-1]/(alphas[k-1]-alphas[k  ])/(alphas[k-1]-alphas[k+1])/(alphas[k-1]-alphas[k+2])
    +(as-alphas[k-1])*(as-alphas[k+1])*(as-alphas[k+2])*xdPqg[k  ]/(alphas[k  ]-alphas[k-1])/(alphas[k  ]-alphas[k+1])/(alphas[k  ]-alphas[k+2])
    +(as-alphas[k-1])*(as-alphas[k  ])*(as-alphas[k+2])*xdPqg[k+1]/(alphas[k+1]-alphas[k-1])/(alphas[k+1]-alphas[k  ])/(alphas[k+1]-alphas[k+2])
    +(as-alphas[k-1])*(as-alphas[k  ])*(as-alphas[k+1])*xdPqg[k+2]/(alphas[k+2]-alphas[k-1])/(alphas[k+2]-alphas[k  ])/(alphas[k+2]-alphas[k+1])
		)/x;

  
  bool useLLp = false; //-->Make this an option
  int RCvar = 0;       //-->Make this an option
  if (opts.order_evol == 1)
    dGgg += as*as * ( HELLx::PLL1(x,hell::nf) - HELLx::mcPgg1LL(x,hell::nf) );
  else if(opts.order_evol == 2)
    {
      dGgg += as*as*as * ( HELLx::PNLL2(x,hell::nf) - CF/CA*HELLx::Pqg2(x,hell::nf,useLLp) - HELLx::mcPgg2NLL(x,hell::nf,useLLp) );
      dGqg += as*as*as *   HELLx::Pqg2(x,hell::nf,useLLp);
    }
  //else if(opts.order_evol == 3)
    //dGqg += as*as*as*as *   HELLx::Pqg3(x,hell::nf,RCvar);
  else if(opts.order_evol == 4)
    {
      dGgg -= as*as*as*as * ( HELLx::PNLL3(x,hell::nf,RCvar) - CF/CA*HELLx::Pqg3(x,hell::nf,useLLp,RCvar) - HELLx::mcPgg3NLL(x,hell::nf,useLLp,RCvar) );
      dGqg -= as*as*as*as *   HELLx::Pqg3(x,hell::nf,useLLp,RCvar);
    }

  HELLx::sqmatrix<double> sg(dGgg, CF/CA*dGgg, dGqg, CF/CA*dGqg);

  fsx[0] = sg.gg();
  fsx[1] = sg.gq();
  fsx[2] = sg.qg();
  fsx[3] = sg.qq();
}
void hell::smallx_eval_2(double x, int i, double as, double fsx[4])
{
  if(as < hell::alphas[0]/2. || as > 4.*hell::alphas[hell::nas-1])
    cout << "HELLx: WARNING: alpha_s=" << as << " out of interpolation range [" << hell::alphas[0] << ", " << hell::alphas[hell::nas-1] << "] for nf=" << hell::nf << endl;

  //k is the index of the closest alphas grid
  int k = 0;
  for (unsigned int i = 1; i < hell::nas; i++)
    {
      if (as < hell::alphas[i]) break;
      k++;
    }
  // the following is needed for cubic interpolation
  if (k==0) k++;
  if (k==hell::nas-1) k -= 2;
  if (k==hell::nas-2) k -= 1;
  
  double xdPgg[hell::nas], xdPqg[hell::nas];
  copy (hell::xdpgg_2+i*hell::nas, hell::xdpgg_2+(i+1)*hell::nas, xdPgg);
  copy (hell::xdpqg_2+i*hell::nas, hell::xdpqg_2+(i+1)*hell::nas, xdPqg);
  
  //cubic interpolation
  double dGgg =(
    +(as-alphas[k  ])*(as-alphas[k+1])*(as-alphas[k+2])*xdPgg[k-1]/(alphas[k-1]-alphas[k  ])/(alphas[k-1]-alphas[k+1])/(alphas[k-1]-alphas[k+2])
    +(as-alphas[k-1])*(as-alphas[k+1])*(as-alphas[k+2])*xdPgg[k  ]/(alphas[k  ]-alphas[k-1])/(alphas[k  ]-alphas[k+1])/(alphas[k  ]-alphas[k+2])
    +(as-alphas[k-1])*(as-alphas[k  ])*(as-alphas[k+2])*xdPgg[k+1]/(alphas[k+1]-alphas[k-1])/(alphas[k+1]-alphas[k  ])/(alphas[k+1]-alphas[k+2])
    +(as-alphas[k-1])*(as-alphas[k  ])*(as-alphas[k+1])*xdPgg[k+2]/(alphas[k+2]-alphas[k-1])/(alphas[k+2]-alphas[k  ])/(alphas[k+2]-alphas[k+1])
		)/x;
  double dGqg =(
    +(as-alphas[k  ])*(as-alphas[k+1])*(as-alphas[k+2])*xdPqg[k-1]/(alphas[k-1]-alphas[k  ])/(alphas[k-1]-alphas[k+1])/(alphas[k-1]-alphas[k+2])
    +(as-alphas[k-1])*(as-alphas[k+1])*(as-alphas[k+2])*xdPqg[k  ]/(alphas[k  ]-alphas[k-1])/(alphas[k  ]-alphas[k+1])/(alphas[k  ]-alphas[k+2])
    +(as-alphas[k-1])*(as-alphas[k  ])*(as-alphas[k+2])*xdPqg[k+1]/(alphas[k+1]-alphas[k-1])/(alphas[k+1]-alphas[k  ])/(alphas[k+1]-alphas[k+2])
    +(as-alphas[k-1])*(as-alphas[k  ])*(as-alphas[k+1])*xdPqg[k+2]/(alphas[k+2]-alphas[k-1])/(alphas[k+2]-alphas[k  ])/(alphas[k+2]-alphas[k+1])
		)/x;

  
  bool useLLp = false; //-->Make this an option
  int RCvar = 0;       //-->Make this an option
  if (opts.order_evol == 1)
    dGgg += as*as * ( HELLx::PLL1(x,hell::nf) - HELLx::mcPgg1LL(x,hell::nf) );
  else if(opts.order_evol == 2)
    {
      dGgg += as*as*as * ( HELLx::PNLL2(x,hell::nf) - CF/CA*HELLx::Pqg2(x,hell::nf,useLLp) - HELLx::mcPgg2NLL(x,hell::nf,useLLp) );
      dGqg += as*as*as *   HELLx::Pqg2(x,hell::nf,useLLp);
    }
  //else if(opts.order_evol == 3)
    //dGqg += as*as*as*as *   HELLx::Pqg3(x,hell::nf,RCvar);
  else if(opts.order_evol == 4)
    {
      dGgg -= as*as*as*as * ( HELLx::PNLL3(x,hell::nf,RCvar) - CF/CA*HELLx::Pqg3(x,hell::nf,useLLp,RCvar) - HELLx::mcPgg3NLL(x,hell::nf,useLLp,RCvar) );
      dGqg -= as*as*as*as *   HELLx::Pqg3(x,hell::nf,useLLp,RCvar);
    }

  HELLx::sqmatrix<double> sg(dGgg, CF/CA*dGgg, dGqg, CF/CA*dGqg);

  fsx[0] = sg.gg();
  fsx[1] = sg.gq();
  fsx[2] = sg.qg();
  fsx[3] = sg.qq();
}

void hell::smallx(double x, double as, double fsx[4])
{
  HELLx::sqmatrix<double> sg = sxtable->DeltaP(as, x, order);
  fsx[0] = sg.gg();
  fsx[1] = sg.gq();
  fsx[2] = sg.qg();
  fsx[3] = sg.qq();
}

HELLx::sqmatrix<complex<double>> DeltaGamma (int i)
{
  complex<double> gg, gq, qg, qq;
  if (i < hell::dim)
    {
      gg = mellinpdf::GMGG[i];
      gq = mellinpdf::GMGQ[i];
      qg = mellinpdf::GMQG[i];
      qq = mellinpdf::GMQQ[i];
    }
  else
    {
      gg = conj(mellinpdf::GMGG[i-hell::dim]);
      gq = conj(mellinpdf::GMGQ[i-hell::dim]);
      qg = conj(mellinpdf::GMQG[i-hell::dim]);
      qq = conj(mellinpdf::GMQQ[i-hell::dim]);
    }
  return HELLx::sqmatrix<complex<double>>( gg, gq, qg, qq );
}
HELLx::sqmatrix<complex<double>> DeltaGamma_1 (int i)
{
  complex<double> gg, gq, qg, qq;
  if (i < hell::dim)
    {
      gg = mellinpdf::GMGG_1[i];
      gq = mellinpdf::GMGQ_1[i];
      qg = mellinpdf::GMQG_1[i];
      qq = mellinpdf::GMQQ_1[i];
    }
  else
    {
      gg = conj(mellinpdf::GMGG_1[i-hell::dim]);
      gq = conj(mellinpdf::GMGQ_1[i-hell::dim]);
      qg = conj(mellinpdf::GMQG_1[i-hell::dim]);
      qq = conj(mellinpdf::GMQQ_1[i-hell::dim]);
    }
  return HELLx::sqmatrix<complex<double>>( gg, gq, qg, qq );
}
HELLx::sqmatrix<complex<double>> DeltaGamma_2 (int i)
{
  complex<double> gg, gq, qg, qq;
  if (i < hell::dim)
    {
      gg = mellinpdf::GMGG_2[i];
      gq = mellinpdf::GMGQ_2[i];
      qg = mellinpdf::GMQG_2[i];
      qq = mellinpdf::GMQQ_2[i];
    }
  else
    {
      gg = conj(mellinpdf::GMGG_2[i-hell::dim]);
      gq = conj(mellinpdf::GMGQ_2[i-hell::dim]);
      qg = conj(mellinpdf::GMQG_2[i-hell::dim]);
      qq = conj(mellinpdf::GMQQ_2[i-hell::dim]);
    }
  return HELLx::sqmatrix<complex<double>>( gg, gq, qg, qq );
}

//Singlet
HELLx::sqmatrix<complex<double>> gamma_LO   (int i) {return HELLx::sqmatrix<complex<double>> (pmom::gamma1gg[i]/M_PI, pmom::gamma1gq[i]/M_PI, 2.*hell::nf*pmom::gamma1qg[i]/M_PI, pmom::gamma1qq[i]/M_PI);}
HELLx::sqmatrix<complex<double>> gamma_NLO  (int i) {return HELLx::sqmatrix<complex<double>> (pmom::gamma2gg[i]/pi2,  pmom::gamma2gq[i]/pi2,  2.*hell::nf*pmom::gamma2qg[i]/pi2, (((pmom::gamma2qq[i]+(hell::nf-1.)*pmom::gamma2qqp[i])+(pmom::gamma2qqb[i]+(hell::nf-1.)*pmom::gamma2qqbp[i]))/pi2));}
HELLx::sqmatrix<complex<double>> gamma_NNLO (int i) {return HELLx::sqmatrix<complex<double>> (pmom::gamma3gg[i]/pi3,  pmom::gamma3gq[i]/pi3,  2.*hell::nf*pmom::gamma3qg[i]/pi3, (((pmom::gamma3qq[i]+(hell::nf-1.)*pmom::gamma3qqp[i])+(pmom::gamma3qqb[i]+(hell::nf-1.)*pmom::gamma3qqbp[i]))/pi3));}
HELLx::sqmatrix<complex<double>> gamma_NNNLO(int i) {return HELLx::sqmatrix<complex<double>> (pmom::gamma4gg[i]/pi4,  pmom::gamma4gq[i]/pi4,  2.*hell::nf*pmom::gamma4qg[i]/pi4, (((pmom::gamma4qq[i]+(hell::nf-1.)*pmom::gamma4qqp[i])+(pmom::gamma4qqb[i]+(hell::nf-1.)*pmom::gamma4qqbp[i]))/pi4));}

//HELLx::sqmatrix<complex<double>> gammaLO_as   (double as, int i) {return ( as*gamma_LO(i) ) / ( -hell::beta0(hell::nf)*as*as );}
//HELLx::sqmatrix<complex<double>> gammaNLO_as  (double as, int i) {return ( as*gamma_LO(i) + as*as*gamma_NLO(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as);}
//HELLx::sqmatrix<complex<double>> gammaNNLO_as (double as, int i) {return ( as*gamma_LO(i) + as*as*gamma_NLO(i) + as*as*as*gamma_NNLO(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as -hell::beta2(hell::nf)*as*as*as*as);}
//HELLx::sqmatrix<complex<double>> gammaNNNLO_as(double as, int i) {return ( as*gamma_LO(i) + as*as*gamma_NLO(i) + as*as*as*gamma_NNLO(i) + as*as*as*as*gamma_NNNLO(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as -hell::beta2(hell::nf)*as*as*as*as -hell::beta3(hell::nf)*as*as*as*as*as);}

HELLx::sqmatrix<complex<double>> gammaLO_as   (double as, int i) {return ( as*gamma_LO(i) ) / evoltrunc::beta(as,0);}
HELLx::sqmatrix<complex<double>> gammaNLO_as  (double as, int i) {return ( as*gamma_LO(i) + as*as*gamma_NLO(i) ) / evoltrunc::beta(as,1);}
HELLx::sqmatrix<complex<double>> gammaNNLO_as (double as, int i) {return ( as*gamma_LO(i) + as*as*gamma_NLO(i) + as*as*as*gamma_NNLO(i) ) / evoltrunc::beta(as,2);}
HELLx::sqmatrix<complex<double>> gammaNNNLO_as(double as, int i) {return ( as*gamma_LO(i) + as*as*gamma_NLO(i) + as*as*as*gamma_NNLO(i) + as*as*as*as*gamma_NNNLO(i) ) / evoltrunc::beta(as,3);}


//HELLx::sqmatrix<complex<double>> gammaLOLL_as    (double as, int i) {return ( as*gamma_LO(i) + DeltaGamma(i) ) / ( -hell::beta0(hell::nf)*as*as );}
//HELLx::sqmatrix<complex<double>> gammaNLONLL_as  (double as, int i) {return ( as*gamma_LO(i) + as*as*gamma_NLO(i) + DeltaGamma(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as);}
//HELLx::sqmatrix<complex<double>> gammaNNLONLL_as (double as, int i) {return ( as*gamma_LO(i) + as*as*gamma_NLO(i) + as*as*as*gamma_NNLO(i) + DeltaGamma(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as -hell::beta2(hell::nf)*as*as*as*as);}
//HELLx::sqmatrix<complex<double>> gammaNNNLONLL_as(double as, int i) {return ( as*gamma_LO(i) + as*as*gamma_NLO(i) + as*as*as*gamma_NNLO(i) + as*as*as*as*gamma_NNNLO(i) + DeltaGamma(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as -hell::beta2(hell::nf)*as*as*as*as -hell::beta3(hell::nf)*as*as*as*as*as);}

HELLx::sqmatrix<complex<double>> gammaLOLL_as    (double as, int i) {return ( as*gamma_LO(i) + DeltaGamma(i) ) / evoltrunc::beta(as,0);}
HELLx::sqmatrix<complex<double>> gammaNLONLL_as  (double as, int i) {return ( as*gamma_LO(i) + as*as*gamma_NLO(i) + DeltaGamma(i) ) / evoltrunc::beta(as,1);}
HELLx::sqmatrix<complex<double>> gammaNNLONLL_as (double as, int i) {return ( as*gamma_LO(i) + as*as*gamma_NLO(i) + as*as*as*gamma_NNLO(i) + DeltaGamma(i) ) / evoltrunc::beta(as,2);}
HELLx::sqmatrix<complex<double>> gammaNNNLONLL_as(double as, int i) {return ( as*gamma_LO(i) + as*as*gamma_NLO(i) + as*as*as*gamma_NNLO(i) + as*as*as*as*gamma_NNNLO(i) + DeltaGamma(i) ) / evoltrunc::beta(as,3);}

HELLx::sqmatrix<complex<double>> gamma_LO_1  (int i) {return HELLx::sqmatrix<complex<double>> (pmom::gamma1gg_1[i]/M_PI, pmom::gamma1gq_1[i]/M_PI, 2.*hell::nf*pmom::gamma1qg_1[i]/M_PI, pmom::gamma1qq_1[i]/M_PI);}
HELLx::sqmatrix<complex<double>> gamma_NLO_1 (int i) {return HELLx::sqmatrix<complex<double>> (pmom::gamma2gg_1[i]/pi2,  pmom::gamma2gq_1[i]/pi2,  2.*hell::nf*pmom::gamma2qg_1[i]/pi2, (((pmom::gamma2qq_1[i]+(hell::nf-1.)*pmom::gamma2qqp_1[i])+(pmom::gamma2qqb_1[i]+(hell::nf-1.)*pmom::gamma2qqbp_1[i]))/pi2));}
HELLx::sqmatrix<complex<double>> gamma_NNLO_1(int i) {return HELLx::sqmatrix<complex<double>> (pmom::gamma3gg_1[i]/pi3,  pmom::gamma3gq_1[i]/pi3,  2.*hell::nf*pmom::gamma3qg_1[i]/pi3, (((pmom::gamma3qq_1[i]+(hell::nf-1.)*pmom::gamma3qqp_1[i])+(pmom::gamma3qqb_1[i]+(hell::nf-1.)*pmom::gamma3qqbp_1[i]))/pi3));}
HELLx::sqmatrix<complex<double>> gammaLO_as_1  (double as, int i) {return ( as*gamma_LO_1(i) ) / ( -hell::beta0(hell::nf)*as*as );}
HELLx::sqmatrix<complex<double>> gammaNLO_as_1 (double as, int i) {return ( as*gamma_LO_1(i) + as*as*gamma_NLO_1(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as);}
HELLx::sqmatrix<complex<double>> gammaNNLO_as_1(double as, int i) {return ( as*gamma_LO_1(i) + as*as*gamma_NLO_1(i) + as*as*as*gamma_NNLO_1(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as -hell::beta2(hell::nf)*as*as*as*as);}
HELLx::sqmatrix<complex<double>> gammaLOLL_as_1   (double as, int i) {return ( as*gamma_LO_1(i) + DeltaGamma_1(i) ) / ( -hell::beta0(hell::nf)*as*as );}
HELLx::sqmatrix<complex<double>> gammaNLONLL_as_1 (double as, int i) {return ( as*gamma_LO_1(i) + as*as*gamma_NLO_1(i) + DeltaGamma_1(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as);}
HELLx::sqmatrix<complex<double>> gammaNNLONLL_as_1(double as, int i) {return ( as*gamma_LO_1(i) + as*as*gamma_NLO_1(i) + as*as*as*gamma_NNLO_1(i) + DeltaGamma_1(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as -hell::beta2(hell::nf)*as*as*as*as);}
HELLx::sqmatrix<complex<double>> gamma_LO_2  (int i) {return HELLx::sqmatrix<complex<double>> (pmom::gamma1gg_2[i]/M_PI, pmom::gamma1gq_2[i]/M_PI, 2.*hell::nf*pmom::gamma1qg_2[i]/M_PI, pmom::gamma1qq_2[i]/M_PI);}
HELLx::sqmatrix<complex<double>> gamma_NLO_2 (int i) {return HELLx::sqmatrix<complex<double>> (pmom::gamma2gg_2[i]/pi2,  pmom::gamma2gq_2[i]/pi2,  2.*hell::nf*pmom::gamma2qg_2[i]/pi2, (((pmom::gamma2qq_2[i]+(hell::nf-1.)*pmom::gamma2qqp_2[i])+(pmom::gamma2qqb_2[i]+(hell::nf-1.)*pmom::gamma2qqbp_2[i]))/pi2));}
HELLx::sqmatrix<complex<double>> gamma_NNLO_2(int i) {return HELLx::sqmatrix<complex<double>> (pmom::gamma3gg_2[i]/pi3,  pmom::gamma3gq_2[i]/pi3,  2.*hell::nf*pmom::gamma3qg_2[i]/pi3, (((pmom::gamma3qq_2[i]+(hell::nf-1.)*pmom::gamma3qqp_2[i])+(pmom::gamma3qqb_2[i]+(hell::nf-1.)*pmom::gamma3qqbp_2[i]))/pi3));}
HELLx::sqmatrix<complex<double>> gammaLO_as_2  (double as, int i) {return ( as*gamma_LO_2(i) ) / ( -hell::beta0(hell::nf)*as*as );}
HELLx::sqmatrix<complex<double>> gammaNLO_as_2 (double as, int i) {return ( as*gamma_LO_2(i) + as*as*gamma_NLO_2(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as);}
HELLx::sqmatrix<complex<double>> gammaNNLO_as_2(double as, int i) {return ( as*gamma_LO_2(i) + as*as*gamma_NLO_2(i) + as*as*as*gamma_NNLO_2(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as -hell::beta2(hell::nf)*as*as*as*as);}
HELLx::sqmatrix<complex<double>> gammaLOLL_as_2   (double as, int i) {return ( as*gamma_LO_2(i) + DeltaGamma_2(i) ) / ( -hell::beta0(hell::nf)*as*as );}
HELLx::sqmatrix<complex<double>> gammaNLONLL_as_2 (double as, int i) {return ( as*gamma_LO_2(i) + as*as*gamma_NLO_2(i) + DeltaGamma_2(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as);}
HELLx::sqmatrix<complex<double>> gammaNNLONLL_as_2(double as, int i) {return ( as*gamma_LO_2(i) + as*as*gamma_NLO_2(i) + as*as*as*gamma_NNLO_2(i) + DeltaGamma_2(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as -hell::beta2(hell::nf)*as*as*as*as);}

//Non Singlet
complex <double> gammaNSp_LO(int i) {return (pmom::gamma1qq[i] + pmom::gamma1qqb[i] - pmom::gamma1qqp[i] - pmom::gamma1qqbp[i])/M_PI;}
complex <double> gammaNSm_LO(int i) {return (pmom::gamma1qq[i] - pmom::gamma1qqb[i] - pmom::gamma1qqp[i] + pmom::gamma1qqbp[i])/M_PI;}
complex <double> gammaNSv_LO(int i) {return (((pmom::gamma1qq[i]+(hell::nf-1.)*pmom::gamma1qqp[i])-(pmom::gamma1qqb[i]+(hell::nf-1.)*pmom::gamma1qqbp[i]))/M_PI);}
complex <double> gammaNSp_NLO(int i) {return (pmom::gamma2qq[i] + pmom::gamma2qqb[i] - pmom::gamma2qqp[i] - pmom::gamma2qqbp[i])/pi2;}
complex <double> gammaNSm_NLO(int i) {return (pmom::gamma2qq[i] - pmom::gamma2qqb[i] - pmom::gamma2qqp[i] + pmom::gamma2qqbp[i])/pi2;}
complex <double> gammaNSv_NLO(int i) {return (((pmom::gamma2qq[i]+(hell::nf-1.)*pmom::gamma2qqp[i])-(pmom::gamma2qqb[i]+(hell::nf-1.)*pmom::gamma2qqbp[i]))/pi2);}
complex <double> gammaNSp_NNLO(int i) {return (pmom::gamma3qq[i] + pmom::gamma3qqb[i] - pmom::gamma3qqp[i] - pmom::gamma3qqbp[i])/pi3;}
complex <double> gammaNSm_NNLO(int i) {return (pmom::gamma3qq[i] - pmom::gamma3qqb[i] - pmom::gamma3qqp[i] + pmom::gamma3qqbp[i])/pi3;}
complex <double> gammaNSv_NNLO(int i) {return (((pmom::gamma3qq[i]+(hell::nf-1.)*pmom::gamma3qqp[i])-(pmom::gamma3qqb[i]+(hell::nf-1.)*pmom::gamma3qqbp[i]))/pi3);}
complex <double> gammaNSp_NNNLO(int i) {return (pmom::gamma4qq[i] + pmom::gamma4qqb[i] - pmom::gamma4qqp[i] - pmom::gamma4qqbp[i])/pi4;}
complex <double> gammaNSm_NNNLO(int i) {return (pmom::gamma4qq[i] - pmom::gamma4qqb[i] - pmom::gamma4qqp[i] + pmom::gamma4qqbp[i])/pi4;}
complex <double> gammaNSv_NNNLO(int i) {return (((pmom::gamma4qq[i]+(hell::nf-1.)*pmom::gamma4qqp[i])-(pmom::gamma4qqb[i]+(hell::nf-1.)*pmom::gamma4qqbp[i]))/pi4);}
complex <double> gammaNSp_LO_1(int i) {return (pmom::gamma1qq_1[i] + pmom::gamma1qqb_1[i] - pmom::gamma1qqp_1[i] - pmom::gamma1qqbp_1[i])/M_PI;}
complex <double> gammaNSm_LO_1(int i) {return (pmom::gamma1qq_1[i] - pmom::gamma1qqb_1[i] - pmom::gamma1qqp_1[i] + pmom::gamma1qqbp_1[i])/M_PI;}
complex <double> gammaNSv_LO_1(int i) {return (((pmom::gamma1qq_1[i]+(hell::nf-1.)*pmom::gamma1qqp_1[i])-(pmom::gamma1qqb_1[i]+(hell::nf-1.)*pmom::gamma1qqbp_1[i]))/M_PI);}
complex <double> gammaNSp_NLO_1(int i) {return (pmom::gamma2qq_1[i] + pmom::gamma2qqb_1[i] - pmom::gamma2qqp_1[i] - pmom::gamma2qqbp_1[i])/pi2;}
complex <double> gammaNSm_NLO_1(int i) {return (pmom::gamma2qq_1[i] - pmom::gamma2qqb_1[i] - pmom::gamma2qqp_1[i] + pmom::gamma2qqbp_1[i])/pi2;}
complex <double> gammaNSv_NLO_1(int i) {return (((pmom::gamma2qq_1[i]+(hell::nf-1.)*pmom::gamma2qqp_1[i])-(pmom::gamma2qqb_1[i]+(hell::nf-1.)*pmom::gamma2qqbp_1[i]))/pi2);}
complex <double> gammaNSp_NNLO_1(int i) {return (pmom::gamma3qq_1[i] + pmom::gamma3qqb_1[i] - pmom::gamma3qqp_1[i] - pmom::gamma3qqbp_1[i])/pi3;}
complex <double> gammaNSm_NNLO_1(int i) {return (pmom::gamma3qq_1[i] - pmom::gamma3qqb_1[i] - pmom::gamma3qqp_1[i] + pmom::gamma3qqbp_1[i])/pi3;}
complex <double> gammaNSv_NNLO_1(int i) {return (((pmom::gamma3qq_1[i]+(hell::nf-1.)*pmom::gamma3qqp_1[i])-(pmom::gamma3qqb_1[i]+(hell::nf-1.)*pmom::gamma3qqbp_1[i]))/pi3);}
complex <double> gammaNSp_NNNLO_1(int i) {return (pmom::gamma4qq_1[i] + pmom::gamma4qqb_1[i] - pmom::gamma4qqp_1[i] - pmom::gamma4qqbp_1[i])/pi4;}
complex <double> gammaNSm_NNNLO_1(int i) {return (pmom::gamma4qq_1[i] - pmom::gamma4qqb_1[i] - pmom::gamma4qqp_1[i] + pmom::gamma4qqbp_1[i])/pi4;}
complex <double> gammaNSv_NNNLO_1(int i) {return (((pmom::gamma4qq_1[i]+(hell::nf-1.)*pmom::gamma4qqp_1[i])-(pmom::gamma4qqb_1[i]+(hell::nf-1.)*pmom::gamma4qqbp_1[i]))/pi4);}
complex <double> gammaNSp_LO_2(int i) {return (pmom::gamma1qq_2[i] + pmom::gamma1qqb_2[i] - pmom::gamma1qqp_2[i] - pmom::gamma1qqbp_2[i])/M_PI;}
complex <double> gammaNSm_LO_2(int i) {return (pmom::gamma1qq_2[i] - pmom::gamma1qqb_2[i] - pmom::gamma1qqp_2[i] + pmom::gamma1qqbp_2[i])/M_PI;}
complex <double> gammaNSv_LO_2(int i) {return (((pmom::gamma1qq_2[i]+(hell::nf-1.)*pmom::gamma1qqp_2[i])-(pmom::gamma1qqb_2[i]+(hell::nf-1.)*pmom::gamma1qqbp_2[i]))/M_PI);}
complex <double> gammaNSp_NLO_2(int i) {return (pmom::gamma2qq_2[i] + pmom::gamma2qqb_2[i] - pmom::gamma2qqp_2[i] - pmom::gamma2qqbp_2[i])/pi2;}
complex <double> gammaNSm_NLO_2(int i) {return (pmom::gamma2qq_2[i] - pmom::gamma2qqb_2[i] - pmom::gamma2qqp_2[i] + pmom::gamma2qqbp_2[i])/pi2;}
complex <double> gammaNSv_NLO_2(int i) {return (((pmom::gamma2qq_2[i]+(hell::nf-1.)*pmom::gamma2qqp_2[i])-(pmom::gamma2qqb_2[i]+(hell::nf-1.)*pmom::gamma2qqbp_2[i]))/pi2);}
complex <double> gammaNSp_NNLO_2(int i) {return (pmom::gamma3qq_2[i] + pmom::gamma3qqb_2[i] - pmom::gamma3qqp_2[i] - pmom::gamma3qqbp_2[i])/pi3;}
complex <double> gammaNSm_NNLO_2(int i) {return (pmom::gamma3qq_2[i] - pmom::gamma3qqb_2[i] - pmom::gamma3qqp_2[i] + pmom::gamma3qqbp_2[i])/pi3;}
complex <double> gammaNSv_NNLO_2(int i) {return (((pmom::gamma3qq_2[i]+(hell::nf-1.)*pmom::gamma3qqp_2[i])-(pmom::gamma3qqb_2[i]+(hell::nf-1.)*pmom::gamma3qqbp_2[i]))/pi3);}
complex <double> gammaNSp_NNNLO_2(int i) {return (pmom::gamma4qq_2[i] + pmom::gamma4qqb_2[i] - pmom::gamma4qqp_2[i] - pmom::gamma4qqbp_2[i])/pi4;}
complex <double> gammaNSm_NNNLO_2(int i) {return (pmom::gamma4qq_2[i] - pmom::gamma4qqb_2[i] - pmom::gamma4qqp_2[i] + pmom::gamma4qqbp_2[i])/pi4;}
complex <double> gammaNSv_NNNLO_2(int i) {return (((pmom::gamma4qq_2[i]+(hell::nf-1.)*pmom::gamma4qqp_2[i])-(pmom::gamma4qqb_2[i]+(hell::nf-1.)*pmom::gamma4qqbp_2[i]))/pi4);}

//complex <double> gammaLO_NSp_as (double as, int i) {return ( as*gammaNSp_LO(i) ) / ( -hell::beta0(hell::nf)*as*as );}
//complex <double> gammaLO_NSm_as (double as, int i) {return ( as*gammaNSm_LO(i) ) / ( -hell::beta0(hell::nf)*as*as );}
//complex <double> gammaLO_NSv_as (double as, int i) {return ( as*gammaNSv_LO(i) ) / ( -hell::beta0(hell::nf)*as*as );}
//complex <double> gammaNLO_NSp_as (double as, int i) {return ( as*gammaNSp_LO(i) + as*as*gammaNSp_NLO(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as);}
//complex <double> gammaNLO_NSm_as (double as, int i) {return ( as*gammaNSm_LO(i) + as*as*gammaNSm_NLO(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as);}
//complex <double> gammaNLO_NSv_as (double as, int i) {return ( as*gammaNSv_LO(i) + as*as*gammaNSv_NLO(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as);}
//complex <double> gammaNNLO_NSp_as (double as, int i) {return ( as*gammaNSp_LO(i) + as*as*gammaNSp_NLO(i) + as*as*as*gammaNSp_NNLO(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as -hell::beta2(hell::nf)*as*as*as*as);}
//complex <double> gammaNNLO_NSm_as (double as, int i) {return ( as*gammaNSm_LO(i) + as*as*gammaNSm_NLO(i) + as*as*as*gammaNSm_NNLO(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as -hell::beta2(hell::nf)*as*as*as*as);}
//complex <double> gammaNNLO_NSv_as (double as, int i) {return ( as*gammaNSv_LO(i) + as*as*gammaNSv_NLO(i) + as*as*as*gammaNSv_NNLO(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as -hell::beta2(hell::nf)*as*as*as*as);}
//complex <double> gammaNNNLO_NSp_as (double as, int i) {return ( as*gammaNSp_LO(i) + as*as*gammaNSp_NLO(i) + as*as*as*gammaNSp_NNLO(i) + as*as*as*as*gammaNSp_NNNLO(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as -hell::beta2(hell::nf)*as*as*as*as -hell::beta3(hell::nf)*as*as*as*as*as);}
//complex <double> gammaNNNLO_NSm_as (double as, int i) {return ( as*gammaNSm_LO(i) + as*as*gammaNSm_NLO(i) + as*as*as*gammaNSm_NNLO(i) + as*as*as*as*gammaNSm_NNNLO(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as -hell::beta2(hell::nf)*as*as*as*as -hell::beta3(hell::nf)*as*as*as*as*as);}
//complex <double> gammaNNNLO_NSv_as (double as, int i) {return ( as*gammaNSv_LO(i) + as*as*gammaNSv_NLO(i) + as*as*as*gammaNSv_NNLO(i) + as*as*as*as*gammaNSv_NNNLO(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as -hell::beta2(hell::nf)*as*as*as*as -hell::beta3(hell::nf)*as*as*as*as*as);}
complex <double> gammaLO_NSp_as_1 (double as, int i) {return ( as*gammaNSp_LO_1(i) ) / ( -hell::beta0(hell::nf)*as*as );}
complex <double> gammaLO_NSm_as_1 (double as, int i) {return ( as*gammaNSm_LO_1(i) ) / ( -hell::beta0(hell::nf)*as*as );}
complex <double> gammaLO_NSv_as_1 (double as, int i) {return ( as*gammaNSv_LO_1(i) ) / ( -hell::beta0(hell::nf)*as*as );}
complex <double> gammaNLO_NSp_as_1 (double as, int i) {return ( as*gammaNSp_LO_1(i) + as*as*gammaNSp_NLO_1(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as);}
complex <double> gammaNLO_NSm_as_1 (double as, int i) {return ( as*gammaNSm_LO_1(i) + as*as*gammaNSm_NLO_1(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as);}
complex <double> gammaNLO_NSv_as_1 (double as, int i) {return ( as*gammaNSv_LO_1(i) + as*as*gammaNSv_NLO_1(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as);}
complex <double> gammaNNLO_NSp_as_1 (double as, int i) {return ( as*gammaNSp_LO_1(i) + as*as*gammaNSp_NLO_1(i) + as*as*as*gammaNSp_NNLO_1(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as -hell::beta2(hell::nf)*as*as*as*as);}
complex <double> gammaNNLO_NSm_as_1 (double as, int i) {return ( as*gammaNSm_LO_1(i) + as*as*gammaNSm_NLO_1(i) + as*as*as*gammaNSm_NNLO_1(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as -hell::beta2(hell::nf)*as*as*as*as);}
complex <double> gammaNNLO_NSv_as_1 (double as, int i) {return ( as*gammaNSv_LO_1(i) + as*as*gammaNSv_NLO_1(i) + as*as*as*gammaNSv_NNLO_1(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as -hell::beta2(hell::nf)*as*as*as*as);}
complex <double> gammaNNNLO_NSp_as_1 (double as, int i) {return ( as*gammaNSp_LO_1(i) + as*as*gammaNSp_NLO_1(i) + as*as*as*gammaNSp_NNLO_1(i) + as*as*as*as*gammaNSp_NNNLO_1(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as -hell::beta2(hell::nf)*as*as*as*as -hell::beta3(hell::nf)*as*as*as*as*as);}
complex <double> gammaNNNLO_NSm_as_1 (double as, int i) {return ( as*gammaNSm_LO_1(i) + as*as*gammaNSm_NLO_1(i) + as*as*as*gammaNSm_NNLO_1(i) + as*as*as*as*gammaNSm_NNNLO_1(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as -hell::beta2(hell::nf)*as*as*as*as -hell::beta3(hell::nf)*as*as*as*as*as);}
complex <double> gammaNNNLO_NSv_as_1 (double as, int i) {return ( as*gammaNSv_LO_1(i) + as*as*gammaNSv_NLO_1(i) + as*as*as*gammaNSv_NNLO_1(i) + as*as*as*as*gammaNSv_NNNLO_1(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as -hell::beta2(hell::nf)*as*as*as*as -hell::beta3(hell::nf)*as*as*as*as*as);}
complex <double> gammaLO_NSp_as_2 (double as, int i) {return ( as*gammaNSp_LO_2(i) ) / ( -hell::beta0(hell::nf)*as*as );}
complex <double> gammaLO_NSm_as_2 (double as, int i) {return ( as*gammaNSm_LO_2(i) ) / ( -hell::beta0(hell::nf)*as*as );}
complex <double> gammaLO_NSv_as_2 (double as, int i) {return ( as*gammaNSv_LO_2(i) ) / ( -hell::beta0(hell::nf)*as*as );}
complex <double> gammaNLO_NSp_as_2 (double as, int i) {return ( as*gammaNSp_LO_2(i) + as*as*gammaNSp_NLO_2(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as);}
complex <double> gammaNLO_NSm_as_2 (double as, int i) {return ( as*gammaNSm_LO_2(i) + as*as*gammaNSm_NLO_2(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as);}
complex <double> gammaNLO_NSv_as_2 (double as, int i) {return ( as*gammaNSv_LO_2(i) + as*as*gammaNSv_NLO_2(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as);}
complex <double> gammaNNLO_NSp_as_2 (double as, int i) {return ( as*gammaNSp_LO_2(i) + as*as*gammaNSp_NLO_2(i) + as*as*as*gammaNSp_NNLO_2(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as -hell::beta2(hell::nf)*as*as*as*as);}
complex <double> gammaNNLO_NSm_as_2 (double as, int i) {return ( as*gammaNSm_LO_2(i) + as*as*gammaNSm_NLO_2(i) + as*as*as*gammaNSm_NNLO_2(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as -hell::beta2(hell::nf)*as*as*as*as);}
complex <double> gammaNNLO_NSv_as_2 (double as, int i) {return ( as*gammaNSv_LO_2(i) + as*as*gammaNSv_NLO_2(i) + as*as*as*gammaNSv_NNLO_2(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as -hell::beta2(hell::nf)*as*as*as*as);}
complex <double> gammaNNNLO_NSp_as_2 (double as, int i) {return ( as*gammaNSp_LO_2(i) + as*as*gammaNSp_NLO_2(i) + as*as*as*gammaNSp_NNLO_2(i) + as*as*as*as*gammaNSp_NNNLO_2(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as -hell::beta2(hell::nf)*as*as*as*as -hell::beta3(hell::nf)*as*as*as*as*as);}
complex <double> gammaNNNLO_NSm_as_2 (double as, int i) {return ( as*gammaNSm_LO_2(i) + as*as*gammaNSm_NLO_2(i) + as*as*as*gammaNSm_NNLO_2(i) + as*as*as*as*gammaNSm_NNNLO_2(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as -hell::beta2(hell::nf)*as*as*as*as -hell::beta3(hell::nf)*as*as*as*as*as);}
complex <double> gammaNNNLO_NSv_as_2 (double as, int i) {return ( as*gammaNSv_LO_2(i) + as*as*gammaNSv_NLO_2(i) + as*as*as*gammaNSv_NNLO_2(i) + as*as*as*as*gammaNSv_NNNLO_2(i) ) / ( -hell::beta0(hell::nf)*as*as -hell::beta1(hell::nf)*as*as*as -hell::beta2(hell::nf)*as*as*as*as -hell::beta3(hell::nf)*as*as*as*as*as);}


complex <double> gammaLO_NSp_as (double as, int i) {return ( as*gammaNSp_LO(i) ) / evoltrunc::beta(as,0);}
complex <double> gammaLO_NSm_as (double as, int i) {return ( as*gammaNSm_LO(i) ) / evoltrunc::beta(as,0);}
complex <double> gammaLO_NSv_as (double as, int i) {return ( as*gammaNSv_LO(i) ) / evoltrunc::beta(as,0);}
complex <double> gammaNLO_NSp_as (double as, int i) {return ( as*gammaNSp_LO(i) + as*as*gammaNSp_NLO(i) ) / evoltrunc::beta(as,1);}
complex <double> gammaNLO_NSm_as (double as, int i) {return ( as*gammaNSm_LO(i) + as*as*gammaNSm_NLO(i) ) / evoltrunc::beta(as,1);}
complex <double> gammaNLO_NSv_as (double as, int i) {return ( as*gammaNSv_LO(i) + as*as*gammaNSv_NLO(i) ) / evoltrunc::beta(as,1);}
complex <double> gammaNNLO_NSp_as (double as, int i) {return ( as*gammaNSp_LO(i) + as*as*gammaNSp_NLO(i) + as*as*as*gammaNSp_NNLO(i) ) / evoltrunc::beta(as,2);}
complex <double> gammaNNLO_NSm_as (double as, int i) {return ( as*gammaNSm_LO(i) + as*as*gammaNSm_NLO(i) + as*as*as*gammaNSm_NNLO(i) ) / evoltrunc::beta(as,2);}
complex <double> gammaNNLO_NSv_as (double as, int i) {return ( as*gammaNSv_LO(i) + as*as*gammaNSv_NLO(i) + as*as*as*gammaNSv_NNLO(i) ) / evoltrunc::beta(as,2);}
complex <double> gammaNNNLO_NSp_as (double as, int i) {return ( as*gammaNSp_LO(i) + as*as*gammaNSp_NLO(i) + as*as*as*gammaNSp_NNLO(i) + as*as*as*as*gammaNSp_NNNLO(i) ) / evoltrunc::beta(as,3);}
complex <double> gammaNNNLO_NSm_as (double as, int i) {return ( as*gammaNSm_LO(i) + as*as*gammaNSm_NLO(i) + as*as*as*gammaNSm_NNLO(i) + as*as*as*as*gammaNSm_NNNLO(i) ) / evoltrunc::beta(as,3);}
complex <double> gammaNNNLO_NSv_as (double as, int i) {return ( as*gammaNSv_LO(i) + as*as*gammaNSv_NLO(i) + as*as*as*gammaNSv_NNLO(i) + as*as*as*as*gammaNSv_NNNLO(i) ) / evoltrunc::beta(as,3);}

//matrix exponential
HELLx::sqmatrix<complex<double>> Uexp(HELLx::sqmatrix<complex<double>> U, double dt)
{
  HELLx::sqmatrix<complex<double>> gam = U;
  complex<double> htr = gam.trace()/2.;
  complex<double> d = gam.det();
  complex<double> gp = htr + sqrt(htr*htr - d);
  complex<double> gm = htr - sqrt(htr*htr - d);
  complex<double> cp = gam.entry21() / (gp - gam.entry22());
  complex<double> cm = gam.entry21() / (gm - gam.entry22());
  HELLx::sqmatrix<complex<double>> Mp(cm, -1., cp*cm, -cp);
  Mp /= (cm-cp);
  HELLx::sqmatrix<complex<double>> Mm(-cp, 1., -cp*cm, cm);
  Mm /= (cm-cp);
  return Mp * exp(dt*gp) + Mm * exp(dt*gm);
}

string sas(double as)
{
  ostringstream os;
  if     (as<0.01) os << "000" << int(1000*as);
  else if(as<0.1 ) os << "00"  << int(1000*as);
  else if(as<1.  ) os << "0"   << int(1000*as);
  else os << int(1000*as);
  return os.str();
}

void hell::init()
{
  nf = resconst::NF;

  if (opts.order_evol <= 1)
    order = HELLx::LO;
  else if (opts.order_evol == 2)
    order = HELLx::NLO;
  else
    order = HELLx::NNLO;

  if (opts.order_evol <= 1)
    logorder = HELLx::LL;
  else
    logorder = HELLx::NLL;

  
  if (opts.mellin1d)
    dim = mellinint::mdim;
  else
    dim = 2*mellinint::mdim;

  if (opts.smallx)
    {
      string datapath;
      char* dytpath = getenv ("DYTPATH");
      if (dytpath != NULL && file_exists(string(dytpath)+"/LL_nf5.info"))
	datapath = string(dytpath)+"/";
      else if (file_exists((string)SHAREDIR+"/LL_nf5.info"))
	datapath = (string)SHAREDIR+"/";
      else if (file_exists("share/dyturbo/LL_nf5.info"))
	datapath = "share/dyturbo/";
      
      if (opts.order_evol == 1)
	sxtable = new HELLx::HELLxnf(hell::nf, HELLx::LL, datapath);
      else if (opts.order_evol >= 2)
	sxtable = new HELLx::HELLxnf(hell::nf, HELLx::NLL, datapath);

      //get list of available alphas tables
      vector <double> alp;
      sxtable->GetAvailableAlphas(alp);
      nas = alp.size();
      alphas = new double [nas];
      copy(alp.begin(), alp.end(), alphas);

      //Load into memory the x tables corresponding to all the available values of alphas
      string basename = "xtable";
      string sord = (opts.order_evol > 1 ? "NLL" : "LL");
      ostringstream filename;
      typename map <int,HELLx::xTableP*>::iterator itxT;  // typename tells the compiler that what follows is a typename, and not a field
      bool quiet = false;
      for(int i = 0; i < nas; i++)
	{
	  filename.str("");  filename.clear();
	  filename << datapath << "/" << sord << "/" << basename << "_" << sord << "_nf" << hell::nf << "_alphas" << sas(hell::alphas[i]) << ".table";
	  xT[i] = new HELLx::xTableP(filename.str(), hell::logorder);
	  xT[i]->SetQuietMode(quiet);
	}
    }
  
//  if (opts.melup <= 1)
//    {
//      allocate();
//      update();
//    }
}

void hell::release()
{
//  if (opts.melup <= 1)
//    free();

  delete sxtable;
}

//Allocate PDF moments at the starting scale
void hell::allocate()
{
  gli  = new complex <double> [2*dim];
  vai  = new complex <double> [2*dim];
  m3i  = new complex <double> [2*dim];
  m8i  = new complex <double> [2*dim];
  m15i = new complex <double> [2*dim];
  m24i = new complex <double> [2*dim];
  sgi  = new complex <double> [2*dim];
  p3i  = new complex <double> [2*dim];
  p8i  = new complex <double> [2*dim];
  p15i = new complex <double> [2*dim];
  p24i = new complex <double> [2*dim];
  cxn = new complex <double> [2*dim];
  if (opts.mellin1d)
    {
      xdpgg = new double[nas*opts.pdfrule];
      xdpqg = new double[nas*opts.pdfrule];
    }
  else
    {
      xdpgg_1 = new double[nas*opts.pdfrule];
      xdpqg_1 = new double[nas*opts.pdfrule];
      xdpgg_2 = new double[nas*opts.pdfrule];
      xdpqg_2 = new double[nas*opts.pdfrule];
    }
}

void hell::allocateusg()
{
  usg = new cmatrix[2*dim];
}

void hell::free()
{
  delete[] gli;
  delete[] vai;
  delete[] m3i;
  delete[] m8i;
  delete[] m15i;
  delete[] m24i;
  delete[] sgi;
  delete[] p3i;
  delete[] p8i;
  delete[] p15i;
  delete[] p24i;
  delete[] cxn;
  if (opts.mellin1d)
    {
      delete[] xdpgg;
      delete[] xdpqg;
    }
  else
    {
      delete[] xdpgg_1;
      delete[] xdpqg_1;
      delete[] xdpgg_2;
      delete[] xdpqg_2;
    }
}
void hell::freeusg()
{
  delete[] usg;
}

//update PDFs at the factorisation scale
void hell::update()
{
  double facscale = scales::fac;
  
  mellinpdf::allocate();
  mellinpdf::evalpdfs(facscale, phasespace::m, phasespace::y);
  mellinpdf::update_mellin();
  mellinpdf::transform();

  complex <double> qp[5];
  complex <double> qm[5];
  complex <double> gl;
  for (int sign = mesq::positive; sign <= mesq::negative; sign++)
    for (int n = 0; n < dim; n++)
      {
	int nn;
	if (opts.mellin1d)
	  nn = n;
	else
	  if (n < mellinint::mdim)
	    nn = n;
	  else
	    nn = n-mellinint::mdim;
	
	if (opts.mellin1d)
	  if (sign == mesq::positive)
	    {
	      gl    = mellinpdf::GL[nn];
	      qp[0] = mellinpdf::UP[nn] + mellinpdf::UB[nn];
	      qp[1] = mellinpdf::DO[nn] + mellinpdf::DB[nn];
	      qp[2] = mellinpdf::ST[nn] + mellinpdf::SB[nn];
	      qp[3] = mellinpdf::CH[nn] + mellinpdf::CB[nn];
	      qp[4] = mellinpdf::BO[nn] + mellinpdf::BB[nn];
	      qm[0] = mellinpdf::UP[nn] - mellinpdf::UB[nn];
	      qm[1] = mellinpdf::DO[nn] - mellinpdf::DB[nn];
	      qm[2] = mellinpdf::ST[nn] - mellinpdf::SB[nn];
	      qm[3] = mellinpdf::CH[nn] - mellinpdf::CB[nn];
	      qm[4] = mellinpdf::BO[nn] - mellinpdf::BB[nn];
	    }
	  else
	    {
	      gl    = conj(mellinpdf::GL[nn]);
	      qp[0] = conj(mellinpdf::UP[nn] + mellinpdf::UB[nn]);
	      qp[1] = conj(mellinpdf::DO[nn] + mellinpdf::DB[nn]);
	      qp[2] = conj(mellinpdf::ST[nn] + mellinpdf::SB[nn]);
	      qp[3] = conj(mellinpdf::CH[nn] + mellinpdf::CB[nn]);
	      qp[4] = conj(mellinpdf::BO[nn] + mellinpdf::BB[nn]);
	      qm[0] = conj(mellinpdf::UP[nn] - mellinpdf::UB[nn]);
	      qm[1] = conj(mellinpdf::DO[nn] - mellinpdf::DB[nn]);
	      qm[2] = conj(mellinpdf::ST[nn] - mellinpdf::SB[nn]);
	      qm[3] = conj(mellinpdf::CH[nn] - mellinpdf::CB[nn]);
	      qm[4] = conj(mellinpdf::BO[nn] - mellinpdf::BB[nn]);
	    }
	else
	  if (n < mellinint::mdim)
	    if (sign == mesq::positive)
	      {
		gl    = mellinpdf::GL_1[nn];
		qp[0] = mellinpdf::UP_1[nn] + mellinpdf::UB_1[nn];
		qp[1] = mellinpdf::DO_1[nn] + mellinpdf::DB_1[nn];
		qp[2] = mellinpdf::ST_1[nn] + mellinpdf::SB_1[nn];
		qp[3] = mellinpdf::CH_1[nn] + mellinpdf::CB_1[nn];
		qp[4] = mellinpdf::BO_1[nn] + mellinpdf::BB_1[nn];
		qm[0] = mellinpdf::UP_1[nn] - mellinpdf::UB_1[nn];
		qm[1] = mellinpdf::DO_1[nn] - mellinpdf::DB_1[nn];
		qm[2] = mellinpdf::ST_1[nn] - mellinpdf::SB_1[nn];
		qm[3] = mellinpdf::CH_1[nn] - mellinpdf::CB_1[nn];
		qm[4] = mellinpdf::BO_1[nn] - mellinpdf::BB_1[nn];
	      }
	    else
	      {
		gl    = conj(mellinpdf::GL_1[nn]);
		qp[0] = conj(mellinpdf::UP_1[nn] + mellinpdf::UB_1[nn]);
		qp[1] = conj(mellinpdf::DO_1[nn] + mellinpdf::DB_1[nn]);
		qp[2] = conj(mellinpdf::ST_1[nn] + mellinpdf::SB_1[nn]);
		qp[3] = conj(mellinpdf::CH_1[nn] + mellinpdf::CB_1[nn]);
		qp[4] = conj(mellinpdf::BO_1[nn] + mellinpdf::BB_1[nn]);
		qm[0] = conj(mellinpdf::UP_1[nn] - mellinpdf::UB_1[nn]);
		qm[1] = conj(mellinpdf::DO_1[nn] - mellinpdf::DB_1[nn]);
		qm[2] = conj(mellinpdf::ST_1[nn] - mellinpdf::SB_1[nn]);
		qm[3] = conj(mellinpdf::CH_1[nn] - mellinpdf::CB_1[nn]);
		qm[4] = conj(mellinpdf::BO_1[nn] - mellinpdf::BB_1[nn]);
	      }
	  else
	    if (sign == mesq::positive)
	      {
		gl    = mellinpdf::GL_2[nn];
		qp[0] = mellinpdf::UP_2[nn] + mellinpdf::UB_2[nn];
		qp[1] = mellinpdf::DO_2[nn] + mellinpdf::DB_2[nn];
		qp[2] = mellinpdf::ST_2[nn] + mellinpdf::SB_2[nn];
		qp[3] = mellinpdf::CH_2[nn] + mellinpdf::CB_2[nn];
		qp[4] = mellinpdf::BO_2[nn] + mellinpdf::BB_2[nn];
		qm[0] = mellinpdf::UP_2[nn] - mellinpdf::UB_2[nn];
		qm[1] = mellinpdf::DO_2[nn] - mellinpdf::DB_2[nn];
		qm[2] = mellinpdf::ST_2[nn] - mellinpdf::SB_2[nn];
		qm[3] = mellinpdf::CH_2[nn] - mellinpdf::CB_2[nn];
		qm[4] = mellinpdf::BO_2[nn] - mellinpdf::BB_2[nn];
	      }
	    else
	      {
		gl    = conj(mellinpdf::GL_2[nn]);
		qp[0] = conj(mellinpdf::UP_2[nn] + mellinpdf::UB_2[nn]);
		qp[1] = conj(mellinpdf::DO_2[nn] + mellinpdf::DB_2[nn]);
		qp[2] = conj(mellinpdf::ST_2[nn] + mellinpdf::SB_2[nn]);
		qp[3] = conj(mellinpdf::CH_2[nn] + mellinpdf::CB_2[nn]);
		qp[4] = conj(mellinpdf::BO_2[nn] + mellinpdf::BB_2[nn]);
		qm[0] = conj(mellinpdf::UP_2[nn] - mellinpdf::UB_2[nn]);
		qm[1] = conj(mellinpdf::DO_2[nn] - mellinpdf::DB_2[nn]);
		qm[2] = conj(mellinpdf::ST_2[nn] - mellinpdf::SB_2[nn]);
		qm[3] = conj(mellinpdf::CH_2[nn] - mellinpdf::CB_2[nn]);
		qm[4] = conj(mellinpdf::BO_2[nn] - mellinpdf::BB_2[nn]);
	      }
	
	//arrays of non-singlet and singlet quark combinations for N_f = 3 (and 4 and 5)
	//defined as in Eq. (2.16) of hep-ph/0408244
	int i = index(n,sign);
	gli[i]  = gl;	
	vai[i]  = qm[0]+qm[1]+qm[2]+qm[3]+qm[4];
	m3i[i]  = qm[0]-qm[1];
	m8i[i]  = qm[0]+qm[1]-2.*qm[2];
	m15i[i] = qm[0]+qm[1]+qm[2]-3.*qm[3];
	m24i[i] = qm[0]+qm[1]+qm[2]+qm[3]-4.*qm[4];
	
	sgi[i]  = qp[0]+qp[1]+qp[2]+qp[3]+qp[4];
	p3i[i]  = qp[0]-qp[1];
	p8i[i]  = qp[0]+qp[1]-2.*qp[2];
	p15i[i] = qp[0]+qp[1]+qp[2]-3.*qp[3];
	p24i[i] = qp[0]+qp[1]+qp[2]+qp[3]-4.*qp[4];
      }

  mellinpdf::updatesx(phasespace::m, phasespace::y);
  if (opts.mellin1d)
    {
      copy(mellinpdf::xdpgg,mellinpdf::xdpgg+nas*opts.pdfrule,hell::xdpgg);
      copy(mellinpdf::xdpqg,mellinpdf::xdpqg+nas*opts.pdfrule,hell::xdpqg);
    }
  else
    {
      copy(mellinpdf::xdpgg_1,mellinpdf::xdpgg_1+nas*opts.pdfrule,hell::xdpgg_1);
      copy(mellinpdf::xdpqg_1,mellinpdf::xdpqg_1+nas*opts.pdfrule,hell::xdpqg_1);
      copy(mellinpdf::xdpgg_2,mellinpdf::xdpgg_2+nas*opts.pdfrule,hell::xdpgg_2);
      copy(mellinpdf::xdpqg_2,mellinpdf::xdpqg_2+nas*opts.pdfrule,hell::xdpqg_2);
    }
  
  mellinpdf::free();

  if (opts.mellin1d)
    for (int sign = mesq::positive; sign <= mesq::negative; sign++)
      for (int i = 0; i < mellinint::mdim; i++)
	{
	  int idx = index(i,sign);
	  if (sign == mesq::positive)
	    cxn[idx] = mellinint::Np[i];
	  else
	    cxn[idx] = mellinint::Nm[i];
	}
  else
    for (int sign = mesq::positive; sign <= mesq::negative; sign++)
      for (int beam = 1; beam <= 2; beam++)
	for (int i = 0; i < mellinint::mdim; i++)
	  {
	    int idx = index(i,beam,sign);
	    if (beam == 1)
	      if (sign == mesq::positive)
		cxn[idx] = mellinint::Np_1[i];
	      else
		cxn[idx] = mellinint::Nm_1[i];
	    else
	      if (sign == mesq::positive)
		cxn[idx] = mellinint::Np_2[i];
	      else
		cxn[idx] = mellinint::Nm_2[i];
	  }
}

void hell::evolve()
{
  //alphas for the evolution
  double asi;
  if (opts.commute)
    asi = pdf::alphas(scales::fac, false);
  else
    asi = pdf::alphas(scales::res, false);
  complex <double> asf = pdfevol::asl*asi;
  
  //Allocate memory for small-x splitting functions and anomalous dimensions
  if (opts.smallx)
    {
      mellinpdf::allocatesx();
      mellinpdf::evalsx(asi, phasespace::m, phasespace::y);
      mellinpdf::update_mellin();
    }

  //Mellin arrays of evolution matrices
  HELLx::sqmatrix<complex<double>> usg[dim*2];
  complex <double> unsp[dim*2];
  complex <double> unsm[dim*2];
  complex <double> unsv[dim*2];
  fill(usg,usg+dim*2, HELLx::sqmatrix<complex<double>>(1.,0.,0.,1.));
  fill(unsp,unsp+dim*2, 1.);
  fill(unsm,unsm+dim*2, 1.);
  fill(unsv,unsv+dim*2, 1.);

  int n,nn,sign;

  //Numerical evolution
  if (opts.iterative == 1)
    {
      HELLx::sqmatrix<complex<double>> logusg[dim*2];
      complex <double> logunsp[dim*2];
      complex <double> logunsm[dim*2];
      complex <double> logunsv[dim*2];
      fill(logusg,logusg+dim*2, HELLx::sqmatrix<complex<double>>(0.,0.,0.,0.));
      fill(logunsp,logunsp+dim*2, 0.);
      fill(logunsm,logunsm+dim*2, 0.);
      fill(logunsv,logunsv+dim*2, 0.);

      //Start alphas loop
      double t = real(asf);
      double t0 = asi;

      //Evaluate singlet evolution operator trough an ordered path integral, opts.nsub is the number of subdivisions for the numerical integration
      double dt = (t-t0)/(opts.nsub+1.);
      HELLx::sqmatrix<complex<double>> gam;
      if (opts.order_evol > 1 || opts.smallx)
	for(int k = 0; k <= opts.nsub; k++)
	  {
	    double as = t0 + (k+0.5)*dt;
	    if (opts.smallx)
	      {
		mellinpdf::evalsx(as, phasespace::m, phasespace::y);
		mellinpdf::gauss_quad_sx();
	      }

	    //Start Mellin loop
	    for (int i = 0; i < dim*2; i++)
	      {
		n    = (i<dim)? i : i-dim;
		sign = (i<dim)? mesq::positive : mesq::negative;

		int nn;
		if (opts.mellin1d)
		  {
		    nn = n;
		    if (opts.smallx)
		      {
			if (opts.order_evol == 1)
			  gam = gammaLOLL_as(as, pmom::index(nn,sign));
			else if (opts.order_evol == 2)
			  gam = gammaNLONLL_as(as, pmom::index(nn,sign));
			else if (opts.order_evol >= 3)
			  gam = gammaNNLONLL_as(as, pmom::index(nn,sign));
			else if (opts.order_evol >= 4)
			  gam = gammaNNNLONLL_as(as, pmom::index(nn,sign));
		      }
		    else
		      {
			if (opts.order_evol == 1)
			  gam = gammaLO_as(as, pmom::index(nn,sign));
			else if (opts.order_evol == 2)
			  gam = gammaNLO_as(as, pmom::index(nn,sign));
			else if (opts.order_evol == 3)
			  gam = gammaNNLO_as(as, pmom::index(nn,sign));
			else if (opts.order_evol >= 4)
			  gam = gammaNNNLO_as(as, pmom::index(nn,sign));
		      }
		  }
		else
		  {
		    if (n < mellinint::mdim)
		      {
			nn = n;
			if (opts.smallx)
			  {
			    if (opts.order_evol == 1)
			      gam = gammaLOLL_as_1(as, pmom::index(nn,sign));
			    else if (opts.order_evol == 2)
			      gam = gammaNLONLL_as_1(as, pmom::index(nn,sign));
			    else if (opts.order_evol >= 3)
			      gam = gammaNNLONLL_as_1(as, pmom::index(nn,sign));
			  }
			else
			  {
			    if (opts.order_evol == 1)
			      gam = gammaLO_as_1(as, pmom::index(nn,sign));
			    else if (opts.order_evol == 2)
			      gam = gammaNLO_as_1(as, pmom::index(nn,sign));
			    else if (opts.order_evol >= 3)
			      gam = gammaNNLO_as_1(as, pmom::index(nn,sign));
			  }
		      }
		    else
		      {
			nn = n-mellinint::mdim;
			if (opts.smallx)
			  {
			    if (opts.order_evol == 1)
			      gam = gammaLOLL_as_2(as, pmom::index(nn,sign));
			    else if (opts.order_evol == 2)
			      gam = gammaNLONLL_as_2(as, pmom::index(nn,sign));
			    else if (opts.order_evol >= 3)
			      gam = gammaNNLONLL_as_2(as, pmom::index(nn,sign));
			  }
			else
			  {
			    if (opts.order_evol == 1)
			      gam = gammaLO_as_2(as, pmom::index(nn,sign));
			    else if (opts.order_evol == 2)
			      gam = gammaNLO_as_2(as, pmom::index(nn,sign));
			    else if (opts.order_evol >= 3)
			      gam = gammaNNLO_as_2(as, pmom::index(nn,sign));
			  }
		      }
		  }
		usg[i] = Uexp(gam, dt) * usg[i];  // Left product (path-ordering)
	      } //end Mellin loop
	  }//end alphas loop

      //Evaluate the non-singlet evolution operator with a Gaussian rule
      int asrule = 8;
      double cc = 0.5*(t+t0);
      double mm = 0.5*(t-t0);
      for (int k = 0; k < asrule; k++)
	{
	  double as = cc+mm*gr::xxx[asrule-1][k];
	  for (int i = 0; i < dim*2; i++)
	    {
	      n    = (i<dim)? i : i-dim;
	      sign = (i<dim)? mesq::positive : mesq::negative;

	      int nn;
	      if (opts.mellin1d)
		{
		  nn = n;
		  if (opts.order_evol == 1 && !opts.smallx)
		    {
		      gam = gammaLO_as(as, pmom::index(nn,sign));
		      logusg[i] = logusg[i]+gam*gr::www[asrule-1][k]*mm;
		    }
		  if (opts.order_evol == 1)
		    {
		      logunsp[i] += gammaLO_NSp_as(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
		      logunsm[i] += gammaLO_NSm_as(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
		      logunsv[i] += gammaLO_NSv_as(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
		    }
		  else if (opts.order_evol == 2)
		    {
		      logunsp[i] += gammaNLO_NSp_as(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
		      logunsm[i] += gammaNLO_NSm_as(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
		      logunsv[i] += gammaNLO_NSv_as(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
		    }
		  else if (opts.order_evol == 3)
		    {
		      logunsp[i] += gammaNNLO_NSp_as(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
		      logunsm[i] += gammaNNLO_NSm_as(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
		      logunsv[i] += gammaNNLO_NSv_as(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
		    }
		  else if (opts.order_evol >= 4)
		    {
		      logunsp[i] += gammaNNNLO_NSp_as(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
		      logunsm[i] += gammaNNNLO_NSm_as(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
		      logunsv[i] += gammaNNNLO_NSv_as(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
		    }
		}
	      else
		{
		  if (n < mellinint::mdim)
		    {
		      nn = n;
		      if (opts.order_evol == 1 && !opts.smallx)
			{
			  gam = gammaLO_as_1(as, pmom::index(nn,sign));
			  logusg[i] = logusg[i]+gam*gr::www[asrule-1][k]*mm;
			}
		      if (opts.order_evol == 1)
			{
			  logunsp[i] += gammaLO_NSp_as_1(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
			  logunsm[i] += gammaLO_NSm_as_1(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
			  logunsv[i] += gammaLO_NSv_as_1(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
			}
		      else if (opts.order_evol == 2)
			{
			  logunsp[i] += gammaNLO_NSp_as_1(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
			  logunsm[i] += gammaNLO_NSm_as_1(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
			  logunsv[i] += gammaNLO_NSv_as_1(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
			}
		      else if (opts.order_evol == 3)
			{
			  logunsp[i] += gammaNNLO_NSp_as_1(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
			  logunsm[i] += gammaNNLO_NSm_as_1(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
			  logunsv[i] += gammaNNLO_NSv_as_1(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
			}
		      else if (opts.order_evol >= 4)
			{
			  logunsp[i] += gammaNNNLO_NSp_as_1(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
			  logunsm[i] += gammaNNNLO_NSm_as_1(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
			  logunsv[i] += gammaNNNLO_NSv_as_1(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
			}
		    }
		  else
		    {
		      nn = n-mellinint::mdim;
		      if (opts.order_evol == 1 && !opts.smallx)
			{
			  gam = gammaLO_as_2(as, pmom::index(nn,sign));
			  logusg[i] = logusg[i]+gam*gr::www[asrule-1][k]*mm;
			}
		      if (opts.order_evol == 1)
			{
			  logunsp[i] += gammaLO_NSp_as_2(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
			  logunsm[i] += gammaLO_NSm_as_2(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
			  logunsv[i] += gammaLO_NSv_as_2(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
			}
		      else if (opts.order_evol == 2)
			{
			  logunsp[i] += gammaNLO_NSp_as_2(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
			  logunsm[i] += gammaNLO_NSm_as_2(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
			  logunsv[i] += gammaNLO_NSv_as_2(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
			}
		      else if (opts.order_evol == 3)
			{
			  logunsp[i] += gammaNNLO_NSp_as_2(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
			  logunsm[i] += gammaNNLO_NSm_as_2(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
			  logunsv[i] += gammaNNLO_NSv_as_2(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
			}
		      else if (opts.order_evol >= 4)
			{
			  logunsp[i] += gammaNNNLO_NSp_as_2(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
			  logunsm[i] += gammaNNNLO_NSm_as_2(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
			  logunsv[i] += gammaNNNLO_NSv_as_2(as, pmom::index(nn,sign))*gr::www[asrule-1][k]*mm;
			}
		    }
	      
		}
	    }
	}

      for (int i = 0; i < dim*2; i++)
	{
	  unsp[i] = exp(logunsp[i]);
	  unsm[i] = exp(logunsm[i]);
	  unsv[i] = exp(logunsv[i]);
	  if (opts.order_evol == 1 && !opts.smallx)
	    usg[i]  = Uexp(logusg[i], 1);
	  //cout << "numerical " << i << "  " << usg[i] << endl;
	}
      //cout << "numerical " << usg[0] << endl;
      //cout << "numerical " << unsp[0] << endl;
      //cout << "numerical " << unsm[0] << endl;
      //cout << "numerical " << unsv[0] << endl;
    }
  else //truncated solution
    {
      double b1 = beta1(hell::nf)/beta0(hell::nf);
      double b2 = beta2(hell::nf)/beta0(hell::nf);
      double b3 = beta3(hell::nf)/beta0(hell::nf);
      HELLx::sqmatrix<complex<double>> I = HELLx::sqmatrix<complex<double>>(1.,0.,0.,1.);

      complex <double> asl = -log(asf/asi);
      
      for (int i = 0; i < dim*2; i++)
	{
	  n    = (i<dim)? i : i-dim;
	  sign = (i<dim)? mesq::positive : mesq::negative;
	  if (opts.mellin1d)
	    nn = n;

	  //compute the LO evolution
	  //complex <double> asint = -log(asf/asi)/beta0(hell::nf);  // integral of the LO as
	  //HELLx::sqmatrix<complex<double>> ULO = Uexp(gamma_LO(pmom::index(nn,sign)), real(asint));
	  //cout << "analytic " << i << "  " <<  usg[i] << endl;

	  /*************  All this part can be cached *******************/
	  //Compute the Rk matrices with Eq. (2.21) of https://arxiv.org/pdf/hep-ph/0408244.pdf
	  HELLx::sqmatrix<complex<double>> R0 = gamma_LO(pmom::index(nn,sign))   /beta0(hell::nf);
	  HELLx::sqmatrix<complex<double>> R1 = gamma_NLO(pmom::index(nn,sign))  /beta0(hell::nf) - b1*R0;
	  HELLx::sqmatrix<complex<double>> R2 = gamma_NNLO(pmom::index(nn,sign)) /beta0(hell::nf) - b1*R1 - b2*R0;
	  HELLx::sqmatrix<complex<double>> R3 = gamma_NNNLO(pmom::index(nn,sign))/beta0(hell::nf) - b1*R2 - b2*R1 - b3*R0;
	  //Compute eigenvalues of the R0 matrix (See Eq. (2.27) of https://arxiv.org/pdf/hep-ph/0408244.pdf)
	  complex<double> htr = R0.trace()/2.;
	  complex<double> d = R0.det();
	  complex<double> rp = htr + sqrt(htr*htr - d); 
	  complex<double> rm = htr - sqrt(htr*htr - d);

	  //Compute eigenvector projections (See Eq. (2.28) of https://arxiv.org/pdf/hep-ph/0408244.pdf)
	  HELLx::sqmatrix<complex<double>> Ep = 1./(rp-rm)*(R0-rm*I);
	  HELLx::sqmatrix<complex<double>> Em = 1./(rm-rp)*(R0-rp*I);

	  //Compute Rktilde and Uk matrices (Eq. (2.25) and (2.31) of https://arxiv.org/pdf/hep-ph/0408244.pdf)
	  HELLx::sqmatrix<complex<double>> R1TL = R1; 
	  HELLx::sqmatrix<complex<double>> U1 = -Em*R1TL*Em/1. - Ep*R1TL*Ep/1. + Ep*R1TL*Em/(rm-rp-1.) + Em*R1TL*Ep/(rp-rm-1.); //Eq. (2.31) of https://arxiv.org/pdf/hep-ph/0408244.pdf
	  HELLx::sqmatrix<complex<double>> R2TL = R2+R1*U1;
	  HELLx::sqmatrix<complex<double>> U2 = -Em*R2TL*Em/2. - Ep*R2TL*Ep/2. + Ep*R2TL*Em/(rm-rp-2.) + Em*R2TL*Ep/(rp-rm-2.); //Eq. (2.31) of https://arxiv.org/pdf/hep-ph/0408244.pdf
	  HELLx::sqmatrix<complex<double>> R3TL = R3+R2*U1+R1*U2;
	  HELLx::sqmatrix<complex<double>> U3 = -Em*R3TL*Em/3. - Ep*R3TL*Ep/3. + Ep*R3TL*Em/(rm-rp-3.) + Em*R3TL*Ep/(rp-rm-3.); //Eq. (2.31) of https://arxiv.org/pdf/hep-ph/0408244.pdf
	  /*************  End caching ************************************/

	  //LO evolution matrix
	  HELLx::sqmatrix<complex<double>> ULO = Em*exp(asl*rm) + Ep*exp(asl*rp); //Eq. (2.29) of https://arxiv.org/pdf/hep-ph/0408244.pdf
	  //cout << "ULO    " << ULO << endl;
	  //cout << "ULO_PG " << ULO_PG << endl;
	  
	  //Eq. (2.24) of https://arxiv.org/pdf/hep-ph/0408244.pdf
	  //LO evolution
	  if (opts.order_evol >= 1)
	    usg[i] = ULO;
      
	  //NLO evolution
	  if (opts.order_evol >= 2)
	    usg[i] += asf*U1*ULO - asi*ULO*U1;

	  //NNLO evolution
	  if (opts.order_evol >= 3)
	    usg[i] += asf*asf*U2*ULO - asf*asi*U1*ULO*U1 + asi*asi*ULO*(U1*U1-U2);

	  //NNNLO evolution
	  if (opts.order_evol >= 4)
	    usg[i] += asf*asf*asf*U3*ULO - asf*asf*asi*U2*ULO*U1 + asf*asi*asi*U1*ULO*(U1*U1-U2)
	      - asi*asi*asi*ULO*(U1*U1*U1-U1*U2-U2*U1+U3); //typo in Eq. (2.24) of https://arxiv.org/pdf/hep-ph/0408244.pdf, see Eq. (34) of https://doi.org/10.1016/0550-3213(94)90076-0

	  //Non singlet
	  //complex <double> asint = asl/beta0(hell::nf);  // integral of the LO as
	  //unsp[i] = exp(gammaNSp_LO(pmom::index(nn,sign))*asint);
	  //unsm[i] = exp(gammaNSm_LO(pmom::index(nn,sign))*asint);
	  //unsv[i] = exp(gammaNSv_LO(pmom::index(nn,sign))*asint);


	  /*************  All this part can be cached *******************/
	  complex <double> R0NSp = gammaNSp_LO(pmom::index(nn,sign))/beta0(hell::nf); //Eq. (2.21) of https://arxiv.org/pdf/hep-ph/0408244.pdf
	  complex <double> R0NSm = gammaNSm_LO(pmom::index(nn,sign))/beta0(hell::nf); //Eq. (2.21) of https://arxiv.org/pdf/hep-ph/0408244.pdf
	  complex <double> R0NSv = gammaNSv_LO(pmom::index(nn,sign))/beta0(hell::nf); //Eq. (2.21) of https://arxiv.org/pdf/hep-ph/0408244.pdf

	  complex <double> R1NSp = gammaNSp_NLO(pmom::index(nn,sign))/beta0(hell::nf)   - b1*R0NSp; //Eq. (2.21) of https://arxiv.org/pdf/hep-ph/0408244.pdf
	  complex <double> R1NSm = gammaNSm_NLO(pmom::index(nn,sign))/beta0(hell::nf)   - b1*R0NSm; //Eq. (2.21) of https://arxiv.org/pdf/hep-ph/0408244.pdf
	  complex <double> R1NSv = gammaNSv_NLO(pmom::index(nn,sign))/beta0(hell::nf)   - b1*R0NSv; //Eq. (2.21) of https://arxiv.org/pdf/hep-ph/0408244.pdf

	  complex <double> R2NSp = gammaNSp_NNLO(pmom::index(nn,sign))/beta0(hell::nf)  - b1*R1NSp - b2*R0NSp; //Eq. (2.21) of https://arxiv.org/pdf/hep-ph/0408244.pdf
	  complex <double> R2NSm = gammaNSm_NNLO(pmom::index(nn,sign))/beta0(hell::nf)  - b1*R1NSm - b2*R0NSm; //Eq. (2.21) of https://arxiv.org/pdf/hep-ph/0408244.pdf
	  complex <double> R2NSv = gammaNSv_NNLO(pmom::index(nn,sign))/beta0(hell::nf)  - b1*R1NSv - b2*R0NSv; //Eq. (2.21) of https://arxiv.org/pdf/hep-ph/0408244.pdf

	  complex <double> R3NSp = gammaNSp_NNNLO(pmom::index(nn,sign))/beta0(hell::nf)  - b1*R2NSp - b2*R1NSp - b3*R0NSp; //Eq. (2.21) of https://arxiv.org/pdf/hep-ph/0408244.pdf
	  complex <double> R3NSm = gammaNSm_NNNLO(pmom::index(nn,sign))/beta0(hell::nf)  - b1*R2NSm - b2*R1NSm - b3*R0NSm; //Eq. (2.21) of https://arxiv.org/pdf/hep-ph/0408244.pdf
	  complex <double> R3NSv = gammaNSv_NNNLO(pmom::index(nn,sign))/beta0(hell::nf)  - b1*R2NSv - b2*R1NSv - b3*R0NSv; //Eq. (2.21) of https://arxiv.org/pdf/hep-ph/0408244.pdf
      
	  complex <double> U1NSp = -R1NSp;
	  complex <double> U1NSm = -R1NSm;
	  complex <double> U1NSv = -R1NSv;

	  complex<double> R2NSTLp = R2NSp+R1NSp*U1NSp;
	  complex<double> R2NSTLm = R2NSm+R1NSm*U1NSm;
	  complex<double> R2NSTLv = R2NSv+R1NSv*U1NSv;
      
	  complex <double> U2NSp = -(R2NSTLp)/2.;
	  complex <double> U2NSm = -(R2NSTLm)/2.;
	  complex <double> U2NSv = -(R2NSTLv)/2.;

	  complex<double> R3NSTLp = R3NSp+R2NSp*U1NSp+R1NSp*U2NSp;
	  complex<double> R3NSTLm = R3NSm+R2NSm*U1NSm+R1NSm*U2NSm;
	  complex<double> R3NSTLv = R3NSv+R2NSv*U1NSv+R1NSv*U2NSv;

	  complex <double> U3NSp = -(R3NSTLp)/3.;
	  complex <double> U3NSm = -(R3NSTLm)/3.;
	  complex <double> U3NSv = -(R3NSTLv)/3.;
	  /*************  End caching ************************************/

	  complex<double> UNSLO = exp(R0NSp*asl);
	  if (opts.order_evol >= 1)
	    {
	      unsp[i] = UNSLO;
	      unsm[i] = UNSLO;
	      unsv[i] = UNSLO;
	    }
      
	  //NLO evolution
	  if (opts.order_evol >= 2)
	    {
	      unsp[i] += (asf-asi)*U1NSp*UNSLO;
	      unsm[i] += (asf-asi)*U1NSm*UNSLO;
	      unsv[i] += (asf-asi)*U1NSv*UNSLO;
	    }

	  //NNLO evolution
	  if (opts.order_evol >= 3)
	    {
	      unsp[i] += asf*asf*U2NSp*UNSLO - asf*asi*U1NSp*U1NSp*UNSLO + asi*asi*(U1NSp*U1NSp-U2NSp)*UNSLO;
	      unsm[i] += asf*asf*U2NSm*UNSLO - asf*asi*U1NSm*U1NSm*UNSLO + asi*asi*(U1NSm*U1NSm-U2NSm)*UNSLO;
	      unsv[i] += asf*asf*U2NSv*UNSLO - asf*asi*U1NSv*U1NSv*UNSLO + asi*asi*(U1NSv*U1NSv-U2NSv)*UNSLO;
	    }

	  //NNNLO evolution
	  if (opts.order_evol >= 4)
	    {
	      unsp[i] += asf*asf*asf*U3NSp*UNSLO - asf*asf*asi*U2NSp*U1NSp*UNSLO + asf*asi*asi*U1NSp*UNSLO*(U1NSp*U1NSp-U2NSp) - asi*asi*asi*UNSLO*(U1NSp*U1NSp*U1NSp-2.*U1NSp*U2NSp+U3NSp);
	      unsm[i] += asf*asf*asf*U3NSm*UNSLO - asf*asf*asi*U2NSm*U1NSm*UNSLO + asf*asi*asi*U1NSm*UNSLO*(U1NSm*U1NSm-U2NSm) - asi*asi*asi*UNSLO*(U1NSm*U1NSm*U1NSm-2.*U1NSm*U2NSm+U3NSm);
	      unsv[i] += asf*asf*asf*U3NSv*UNSLO - asf*asf*asi*U2NSv*U1NSv*UNSLO + asf*asi*asi*U1NSv*UNSLO*(U1NSv*U1NSv-U2NSv) - asi*asi*asi*UNSLO*(U1NSv*U1NSv*U1NSv-2.*U1NSv*U2NSv+U3NSv);
	    }

	  
	}
      //cout << "analytic usg " << usg[0] << endl;
      //cout << "analytic unsp " << unsp[0] << endl;
      //cout << "analytic unsm " << unsm[0] << endl;
      //cout << "analytic unsv " << unsv[0] << endl;
    }
  //Loop on Mellin moments
  //cout << endl;
  //cout << "alphas " << asi << "  " << asf << endl;
  for (int i = 0; i < dim*2; i++)
    {
      n    = (i<dim)? i : i-dim;
      sign = (i<dim)? mesq::positive : mesq::negative;

      //PDFs at the starting scale
      HELLx::vec2<complex<double>> f0(gli[i],sgi[i]); //(gluon,singlet)

      //Evolved PDFs

      //singlet
      HELLx::vec2<complex<double>> f1 = usg[i]*f0;
      complex <double> glf = f1.entry1();
      complex <double> sgf = f1.entry2();
      
      //non singlet
      complex <double> vaf  = unsv[i]*vai[i];
      complex <double> m3f  = unsm[i]*m3i[i];
      complex <double> m8f  = unsm[i]*m8i[i];
      complex <double> m15f = unsm[i]*m15i[i];
      complex <double> m24f = unsm[i]*m24i[i];
      complex <double> p3f  = unsp[i]*p3i[i];
      complex <double> p8f  = unsp[i]*p8i[i];
      complex <double> p15f = unsp[i]*p15i[i];
      complex <double> p24f = unsp[i]*p24i[i];
      
      //Flavour decomposition of the minus sector:  valence distributions
      complex <double> bv = (vaf - m24f) * 0.2;
      complex <double> cv = bv + (m24f - m15f) * 0.25;
      complex <double> sv = cv + (m15f - m8f) /3.;
      complex <double> dv = sv + (m8f - m3f) * 0.5;
      complex <double> uv = sv + (m8f + m3f) * 0.5;
      //Flavour decomposition of the plus sector: antiquark distributions
      complex <double> bp = (sgf - p24f) * 0.2;
      complex <double> cp = bp + (p24f - p15f) * 0.25;
      complex <double> sp = cp + (p15f - p8f) /3.;
      complex <double> dp = sp + (p8f - p3f) * 0.5;
      complex <double> up = sp + (p8f + p3f) * 0.5;

      complex <double> fx[11];
      fx[u ] = 0.5 * (uv + up);
      fx[d ] = 0.5 * (dv + dp);
      fx[s ] = 0.5 * (sv + sp);
      fx[c ] = 0.5 * (cv + cp);
      fx[b ] = 0.5 * (bv + bp);
      fx[g ] = glf;
      fx[ub] = 0.5 * (up - uv); 
      fx[db] = 0.5 * (dp - dv);
      fx[sb] = 0.5 * (sp - sv);
      fx[cb] = 0.5 * (cp - cv);
      fx[bb] = 0.5 * (bp - bv);
      
      //cout << "hell " << asf << "  "
      //	   << i << "  " << cxn[i] << "  "
      //	   << fx[g] << "  " << uv << "  " << dv << "  " << fx[ub] << "  " << fx[db] << "  " << fx[s] << "  " << fx[c] << "  " << fx[b]  << endl;

      if (opts.mellin1d)
	pdfevol::storemoments(n, sign, fx);
      else
	if (n < mellinint::mdim)
	  pdfevol::storemoments_1(n, sign, fx);
	else
	  pdfevol::storemoments_2(n-mellinint::mdim, sign, fx);
    }

  //evolnum::calculate();
  //for (int i = 0; i < mellinint::mdim; i++)
  //  {
  //    //int i = 0;
  //    //cout << "direct " << fabs(pdfevol::bstarscale) << "  "
  //    cout << "direct " << real(pdfevol::mubstartilde) << "  "
  //	   << i << "  " << mellinint::Np[i] << "  "
  //	   << pdfevol::fx1[i*11+(0+MAXNF )] << "  "
  //	   << pdfevol::fx1[i*11+(1+MAXNF )]-pdfevol::fx1[i*11+(-1+MAXNF)] << "  "
  //	   << pdfevol::fx1[i*11+(2+MAXNF )]-pdfevol::fx1[i*11+(-2+MAXNF)] << "  "
  //	   << pdfevol::fx1[i*11+(-1+MAXNF)] << "  "
  //	   << pdfevol::fx1[i*11+(-2+MAXNF)] << "  "
  //	   << pdfevol::fx1[i*11+(3+MAXNF )] << "  "
  //	   << pdfevol::fx1[i*11+(4+MAXNF )] << "  "
  //	   << pdfevol::fx1[i*11+(5+MAXNF )] << endl;
  //  }
  
  if (opts.smallx)
    mellinpdf::freesx();
  return;
}



void hell::calcusg(double asi, complex <double> asf, int order)
{
  //Allocate memory for small-x splitting functions and anomalous dimensions
  if (opts.smallx)
    {
      mellinpdf::allocatesx();
      mellinpdf::evalsx(asi, phasespace::m, phasespace::y);
      mellinpdf::update_mellin();
    }

  //Mellin arrays of evolution matrices
  //HELLx::sqmatrix<complex<double>> usg[dim*2]; //--> allocate usg dynamically
  fill(usg,usg+dim*2, HELLx::sqmatrix<complex<double>>(1.,0.,0.,1.));

  int n,nn,sign;

  //Numerical evolution

  //Start alphas loop
  double t = real(asf);
  double t0 = asi;

  //Evaluate singlet evolution operator trough an ordered path integral, opts.nsub is the number of subdivisions for the numerical integration
  double dt = (t-t0)/(opts.nsub+1.);
  HELLx::sqmatrix<complex<double>> gam;
  for(int k = 0; k <= opts.nsub; k++)
    {
      double as = t0 + (k+0.5)*dt;
      if (opts.smallx)
	{
	  mellinpdf::evalsx(as, phasespace::m, phasespace::y);
	  mellinpdf::gauss_quad_sx();
	}

      //Start Mellin loop
      for (int i = 0; i < dim*2; i++)
	{
	  n    = (i<dim)? i : i-dim;
	  sign = (i<dim)? mesq::positive : mesq::negative;

	  int nn;
	  if (opts.mellin1d)
	    {
	      nn = n;
	      if (opts.smallx)
		{
		  if (opts.order_evol == 1)
		    gam = gammaLOLL_as(as, pmom::index(nn,sign));
		  else if (opts.order_evol == 2)
		    gam = gammaNLONLL_as(as, pmom::index(nn,sign));
		  else if (opts.order_evol >= 3)
		    gam = gammaNNLONLL_as(as, pmom::index(nn,sign));
		  else if (opts.order_evol >= 4)
		    gam = gammaNNNLONLL_as(as, pmom::index(nn,sign));
		}
	    }
	  else
	    {
	      if (n < mellinint::mdim)
		{
		  nn = n;
		  if (opts.smallx)
		    {
		      if (opts.order_evol == 1)
			gam = gammaLOLL_as_1(as, pmom::index(nn,sign));
		      else if (opts.order_evol == 2)
			gam = gammaNLONLL_as_1(as, pmom::index(nn,sign));
		      else if (opts.order_evol >= 3)
			gam = gammaNNLONLL_as_1(as, pmom::index(nn,sign));
		    }
		}
	      else
		{
		  nn = n-mellinint::mdim;
		  if (opts.smallx)
		    {
		      if (opts.order_evol == 1)
			gam = gammaLOLL_as_2(as, pmom::index(nn,sign));
		      else if (opts.order_evol == 2)
			gam = gammaNLONLL_as_2(as, pmom::index(nn,sign));
		      else if (opts.order_evol >= 3)
			gam = gammaNNLONLL_as_2(as, pmom::index(nn,sign));
		    }
		}
	    }
	  usg[i] = Uexp(gam, dt) * usg[i];  // Left product (path-ordering)
	} //end Mellin loop
    }//end alphas loop

  if (opts.smallx)
    mellinpdf::freesx();
  return;
}
