#include "npff.h"
#include "settings.h"
#include "resconst.h"
#include "constants.h"
#include "pdfevol.h"
#include "resint.h"
//#include <LHAPDF/LHAPDF.h>
#include "pdf.h"
#include "blim.h"
#include "gaussrules.h"
#include "specialfunctions_interface.h"
#include <iostream>

using namespace constants;

complex <double> npff::uvff;
complex <double> npff::usff;
complex <double> npff::dvff;
complex <double> npff::dsff;
complex <double> npff::ssff;
complex <double> npff::chff;
complex <double> npff::boff;
complex <double> npff::glff;

void npff::init()
{
  //calculate g0(blim) according to Eq. (81) of https://arxiv.org/pdf/1412.3820.pdf
  int rule = 100;

  double q0 = resconst::b0/opts.bref;
  double q1 = resconst::b0/opts.blim;
  double cc = 0.5*(q1+q0);
  double mm = 0.5*(q1-q0);
  double res = 0;
  for (int i = 0; i < rule; i++)
    {
      double mu = cc+mm*gr::xxx[rule-1][i];
      double fac = 2.*constants::CF/M_PI*gr::www[rule-1][i]*mm;
      res += fac*pdf::alphas(mu,opts.cmw)/mu;
      //cout << mu << "  " << fac << "  " << pdf::alphas(mu,opts.cmw) << endl;
    }

  opts.g0 += res;
  //cout << " g0 is now " << opts.g0 << endl;
}

complex <double> npff::S(complex <double> b, double m, double x1, double x2)
{
  complex <double> ff = 1;

  complex <double> bb;

  //3 possible choices
  //bb = b;
  //bb = real(b);
  //bb = abs(b);
  
  //Gaussian form
  //BLNY (blim = 0.5)
  //g1 = 0.21;  //g2 = 0.68;  //g3 = -0.13;  //Q0 = 3.2;
  //Konychev Nadolsky (blim = 1.5)
  //g1 = 0.20;  //g2 = 0.19;  //g3 = -0.03;  //Q0 = 3.2;  
  //M. Hirai, H. Kawamura, K. Tanaka, https://inspirehep.net/record/1229722 (minimal prescription, MSTW08)
  //g1 = 0.330;  //g2 = 0.066;  //g3 = 0.;  //Q0 = 2.*1.3;
  //g1 = 0.50;   //g2 = 2.0;  //g3 = 0.;  //Q0 = 50;
  if (opts.npff == 0)
    {
      double gg1 = opts.g1;
      if (opts.g1x != 0)
	gg1 += opts.g1x/2./beta_euler(opts.g1a+1.,opts.g1b+1.) * ( pow(x1,opts.g1a)*pow(1.-x1,opts.g1b)  + pow(x2,opts.g1a)*pow(1.-x2,opts.g1b));
      double gg2 = opts.g2 * log(m/opts.Q0);
      double gg3 = opts.g3*log(100*x1*x2);

      double g = max(0.,gg1+gg2+gg3);

      ff = exp(-g * pow(abs(b),2));

      //Quartic term
      //if (opts.q > 0)
      ff *= exp(-abs(opts.q) * pow(abs(b),4));
      /*
      //ff = exp(-(g1 + opts.g2 * log(m/opts.Q0) + opts.g3*log(100*x1*x2)) * pow(b,2));
      //ff = exp(-(g1 + opts.g2 * log(m/opts.Q0) + opts.g3*log(100*x1*x2)) * pow(real(b),2));
      ff = exp(-(g1 + opts.g2 * log(m/opts.Q0) + opts.g3*log(100*x1*x2)) * pow(abs(b),2));
      */    
    }

  //Exponential form
  if (opts.npff == 1)
    ff = exp(-opts.e*b);

  //Collins, Rogers, large bT form (https://arxiv.org/abs/1412.3820 Eq.(39), (79))
  //g0 = 0.3; //Q0 = 1.6;
  if (opts.npff == 2)
    {
      //double gk = opts.g0 * (1. - exp(- (CF*pdf::extalphas(real(pdfevol::mubstar))*pow(real(b),2))/(M_PI*opts.g0*pow(blim::sudakov,2))));
      double gk = opts.g0 * (1. - exp(- (CF*pdf::alphas(real(pdfevol::mubstar),opts.cmw)*pow(real(b),2))/(M_PI*opts.g0*pow(blim::sudakov,2))));
      ff = exp(-gk*log(pow(m/opts.Q0,2)));

      //Logaritmic term as in Eq. (13) of https://arxiv.org/pdf/1406.3073.pdf (See also Eq.(67) of https://arxiv.org/pdf/1412.3820.pdf and Eq. (10a) of https://inspirehep.net/literature/217570)
      //if (opts.g2 != 0)
      //ff *= exp(-opts.g2*log(1.+pow(real(b)/blim::sudakov,2))*log(m/opts.Q0));
      
      //Gaussian term
      if (opts.g1 != 0)
	//ff *= exp(-opts.g1                               * pow(abs(b),2));
        ff *= exp(-opts.g1/sqrt(1.+max(0.,opts.lambda)*pow(b,2)) * pow(abs(b),2));
      
      //Quartic term
      //if (opts.q != 0)
      //ff *= exp(-opts.q * pow(abs(b),4));
      
      //if (opts.q != 0)
      //ff *= exp(-opts.q * pow(abs(b),4) - max(0.,-opts.q) * pow(abs(b),6));
      
      double lambdaq = 1;
      if (opts.q != 0)
	ff *= exp(- (opts.q/abs(opts.q))* pow(lambdaq,4)  * (1. - exp(- abs(opts.q)*pow(real(b),4)/pow(lambdaq,4))));

      //if (opts.q != 0)
      //ff *= exp(-opts.q * pow(abs(b),4)/(1.+pow(real(b),4)/pow(lambdaq,4)));

      //Sextic term
      if (opts.sx > 0)
	ff *= exp(-opts.sx * pow(abs(b),6));

      //cout << "Collins-Rogers " << " mubstar " << pdfevol::mubstar << " alphas " << pdf::alphas(real(pdfevol::mubstar),opts.cmw) << " lha alphas " << pdf::extalphas(real(pdfevol::mubstar)) << endl;
    }

  //Dispersive approach (DMW) https://arxiv.org/abs/hep-ph/9512336  https://iopscience.iop.org/article/10.1088/1126-6708/1999/07/012/pdf
  //# 3: Dokshitzer-Marchesini-Webber    npff(b) = exp(-(a2 * (log(m)+1.) - a2p) * b^2/2);
  //# Dokshitzer-Marchesini-Webber options (npff = 3)
  //a2 = 0.2
  //a2p = 0
  //Also the N-independent part is included in expc
  if (opts.npff == 3)
    ff = exp(-(opts.a2 * (2.*log(m)+1.) - opts.a2p) * pow(real(b),2)/2. * 2.); //one factor for each leg -> multiply by 2
  //ff = 1.;

  //Pavia19 fit
  if (opts.npff == 4)
    {
      //From Table 5 of https://arxiv.org/pdf/1912.07550.pdf
      double lambda = opts.lambda; //0.580;
      double g2 = opts.g2A; //0.036;   //ln(m)-dependent quadratic (Gaussian) term
      double g2B = opts.g2B; //0.012;  //ln(m)-dependent quartic term
      double N1 = opts.NA; //0.625;
      double sigma = opts.sigmaA; //0.370;
      double alpha = opts.alphaA; //0.205;
      double N1B = opts.NB; //0.044;
      double sigmaB = opts.sigmaB; //0.356;
      double alphaB = opts.alphaB; //0.069;
      double Q0 = 1;
      
      //https://arxiv.org/pdf/1912.07550.pdf Eq. (2.39) and (2.40)
      double g1_1  = N1  / (x1*sigma)  * exp(-1./(2.*pow(sigma,2))  * pow(log(x1/alpha),2) );
      double g1B_1 = N1B / (x1*sigmaB) * exp(-1./(2.*pow(sigmaB,2)) * pow(log(x1/alphaB),2) );

      double g1_2  = N1  / (x2*sigma)  * exp(-1./(2.*pow(sigma,2))  * pow(log(x2/alpha),2) );
      double g1B_2 = N1B / (x2*sigmaB) * exp(-1./(2.*pow(sigmaB,2)) * pow(log(x2/alphaB),2) );

      double four = 4.;
      complex <double> ff_1 = (1.-lambda)/(1.+g1_1*pow(b,2)/four) + lambda*exp(-g1B_1*pow(b,2)/four);
      complex <double> ff_2 = (1.-lambda)/(1.+g1_2*pow(b,2)/four) + lambda*exp(-g1B_2*pow(b,2)/four);
      //ff = ff_1*ff_2*pow(exp(-(g2+g2B*pow(b,2)) *log(pow(m/Q0,2)*pow(b,2)/4.)),2);
      ff = ff_1*ff_2*pow(exp(-    (g2+g2B*4.*pow(abs(b),2)/four) * log(pow(m/Q0,2))* pow(abs(b),2)/four),2);
      //ff =     exp(- 2.*(g2+g2B*pow(abs(b),2)) * log(m*m/Q0/Q0  )* pow(abs(b),2)/4.);

    }

  //SV19 fit
  if (opts.npff == 5)
    {
      
      /*
      //x1 and x2 defined as in https://arxiv.org/pdf/1912.06532.pdf Eq.(2.42)
      double exppy = exp(phasespace::y);
      double expmy = 1./exppy;
      tau = sqrt(m*m+pow(phasespace::qt,2))/opts.sroot;
      double xx1 = tau*exppy;
      double xx2 = tau*expmy;
      */
      double xx1 = x1;
      double xx2 = x2;

      //https://arxiv.org/pdf/1912.06532.pdf Table 9
      double lambda1 = 0.224;
      double lambda2 = 9.24;
      double lambda3 = 375.;
      double lambda4 = 2.15;
      double lambda5 = -4.97;
      
      //https://arxiv.org/pdf/1912.06532.pdf Eq.(2.86)
      complex <double> ff_1 = exp((-lambda1*(1.-xx1) + lambda2*xx1+ xx1*(1.-xx1)*lambda5) / sqrt(1.+lambda3*pow(xx1,lambda4)*pow(b,2)) * pow(b,2));
      complex <double> ff_2 = exp((-lambda1*(1.-xx2) + lambda2*xx2+ xx2*(1.-xx2)*lambda5) / sqrt(1.+lambda3*pow(xx2,lambda4)*pow(b,2)) * pow(b,2));
      ff = ff_1*ff_2;
    }

  if (opts.npff == 6)
    ff = exp( -opts.g1/sqrt(1.+max(0.,opts.lambda)*pow(b,2)) * pow(b,2));

  //Reduce the dependence on bmax using https://arxiv.org/pdf/1412.3820.pdf Eq. (67) and Nucl.Phys. B284, 253 (1987), Eq. (10a)
  if (opts.npff == 7)
    {
      //ff = exp(opts.lambda * CF*pdf::alphas(real(pdfevol::mubstar),opts.cmw)/M_PI*log(1.+pow(real(b)/blim::sudakov,2)));
      //ff = exp(-8.*CF*pdf::alphas(real(pdfevol::mubstar),opts.cmw)/M_PI*( log(1.+pow(real(b)/blim::sudakov,2)) - log(1.+pow(real(b)/3.0,2)) ) );
      //ff = exp(-2.* CF*pdf::alphas(real(pdfevol::mubstar),opts.cmw)/M_PI*log(1.+pow(real(b)/blim::sudakov,2)));
      ff = exp(opts.lambda*log(1.+pow(real(b)/blim::sudakov,2)));

      double gk = opts.g0 * (1. - exp(- (CF*pdf::alphas(real(pdfevol::mubstar),opts.cmw)*pow(real(b),2))/(M_PI*opts.g0*pow(blim::sudakov,2))));
      ff *= exp(-gk*log(pow(m/opts.Q0,2)));

      //Gaussian term
      if (opts.g1 > 0)
	ff *= exp(-opts.g1 * pow(abs(b),2));

	
      //Quartic term
      if (opts.q > 0)
	ff *= exp(-opts.q * pow(abs(b),4));

      //cout << "Collins-Rogers " << " mubstar " << pdfevol::mubstar << " alphas " << pdf::alphas(real(pdfevol::mubstar),opts.cmw) << " lha alphas " << pdf::extalphas(real(pdfevol::mubstar)) << endl;
    }
  
  
  //Additional flavor dependent form factors
  if (opts.flavour_kt)
    {
      uvff = exp(-opts.g1_uv*pow(b,2));// * ff;
      usff = exp(-opts.g1_us*pow(b,2));// * ff;
      dvff = exp(-opts.g1_dv*pow(b,2));// * ff;
      dsff = exp(-opts.g1_ds*pow(b,2));// * ff;
      ssff = exp(-opts.g1_ss*pow(b,2));// * ff;
      chff = exp(-opts.g1_ch*pow(b,2));// * ff;
      boff = exp(-opts.g1_bo*pow(b,2));// * ff;
      glff = exp(-opts.g1_gl*pow(b,2));// * ff;
    }
  
  //universal form factor
  return ff;
}
